#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on April 19 2017

@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

from ...pyqtgraphCore.Qt import QtCore, QtGui, QtWidgets
from ...pyqtgraphCore.imageview import ImageView
import numpy as np
# import abc
# import multiprocessing


class ViewerUtils(QtCore.QObject):
    """
    Some utility functions for interfacing viewer.core.ViewerWorkEnv with the pyqtgraphCore.ImageView widget
    """
    def __init__(self, viewer_reference: ImageView):
        """
        :type viewer_reference: ImageView
        """
        QtCore.QObject.__init__(self)
        # assert isinstance(viewer_reference, ImageView)
        self.viewer = viewer_reference  #: reference to the pyqtgraph ImageView widget instance (viewer)
        self.work_env = self.viewer.workEnv  #: ViewerWorkEnv instance
        self.roi_manager = None

    def update_workEnv(self):
        """
        Update the ImageView widget with the ViewerWorkEnv
        """

        self.viewer.status_bar_label.showMessage('Updating work environment, please wait...')
        self.viewer.parent().show()

        if self.viewer.workEnv.imgdata.ndim == 4:
            self.viewer.set_zlevel_ui_visible(True)
            self.viewer.ui.verticalSliderZLevel.setMaximum(self.viewer.workEnv.imgdata.z_max)
            self.viewer.ui.verticalSliderZLevel.setValue(0)
            self.viewer.ui.spinBoxZLevel.setMaximum(self.viewer.workEnv.imgdata.z_max)
        else:
            self.viewer.set_zlevel_ui_visible(False)

        self.viewer.setImage(self.viewer.workEnv.imgdata.seq.T, pos=(0, 0), scale=(1, 1),
                                   xvals=np.linspace(1, self.viewer.workEnv.imgdata.seq.T.shape[0],
                                                     self.viewer.workEnv.imgdata.seq.T.shape[0]))

        self.viewer.workEnv.roi_manager = self.viewer.parent().roi_manager.manager
        if self.viewer.workEnv.stim_maps is not None:
            # print(self.viewer.workEnv.stim_maps)
            smm = self.viewer.parent().run_module('stimulus_mapping')
            smm.set_all_data(self.viewer.workEnv.stim_maps)
            try:
                smm.set_timeline(1)
            except:
                pass

        self.viewer.sig_workEnv_changed.emit(self.viewer.workEnv)

        self.viewer.status_bar_label.showMessage('Finished updating work environment.')

    def set_statusbar(self, msg):
        """
        Set the status bar message in the viewer window.

        :param msg: text to display in the status bar
        """
        self.viewer.status_bar_label.showMessage(msg)

    def enable_ui(self, b):
        self.viewer.ui.splitter.setEnabled(b)

    def discard_workEnv(self, clear_sample_id=False):
        """
        Ask the user if they want to discard their work environment. If Yes, calls _clear_workEnv()
        """

        self.viewer.parent().show()

        if self.viewer.workEnv.isEmpty:
            return True

        changes = self.viewer.workEnv.changed_items

        if (self.viewer.workEnv.saved is False) and (QtWidgets.QMessageBox.warning(self.viewer, 'Warning!',
                         'You have the following unsaved changes in your work environment. '
                         'Would you like to discard them and continue?\n\n' +
                             ' > '.join(sorted(set(changes), key=changes.index)),
                         QtWidgets.QMessageBox.Yes,QtWidgets.QMessageBox.No)) == QtWidgets.QMessageBox.No:
            self.viewer.status_bar_label.showMessage('Work environment left unchanged.')

            return False
        self._clear_workEnv(clear_sample_id)
        return True

    def _clear_workEnv(self, clear_sample_id=False):
        """
        Cleanup of the ViewerWorkEnv and ImageView widget
        """

        if self.viewer.workEnv.roi_manager is not None:
            self.viewer.workEnv.roi_manager.clear()
        # re-initialize ROI and curve lists
        self.viewer.workEnv.clear()
        self.viewer.clear()
        self.viewer.workEnv.saved = True
        self.viewer.ui.label_curr_img_seq_name.setText('EMPTY')
        self.viewer.set_zlevel_ui_visible(False)
        self.viewer.status_bar_label.showMessage('Work environment cleared.')

    def workEnv_changed(self, element=None):
        if self.viewer.workEnv is not None:
            self.viewer.workEnv.saved = False

        if element is not None:
            self.viewer.workEnv.changed_items.append(element)



"""
Module for importing data from MESc HDF5 files created from Femtonics microscopes.
"""

from .pytemplates.femtonics_mesc_template import *
import h5py
from datetime import datetime
import os
from ...common.qdialogs import use_open_file_dialog, present_exceptions
from ..core import ViewerUtils, ViewerWorkEnv
from ..core.data_types import ImgData
from typing import List, Union
from re import sub as regex_sub
from ...pyqtgraphCore import PlotWidget
import numpy as np
from cv2 import bitwise_not


def ascii_to_str(array: np.ndarray):
    """
    Get a string from an array of ascii integers

    :param array: array of integers representing ascii chars
    :return: string represented by the array
    """
    return ''.join(
        map(
            chr, filter(None, array)
        )
    )


# Navigator object which helps control the
# user's file navigation through the .mesc file
class MEScNavigator(QtWidgets.QWidget):
    # emitted every time the hdf path changes
    # used by ModuleGUI() to determine if Import Image button should be enabled
    sig_hpath_changed: QtCore.pyqtSignal = QtCore.pyqtSignal(dict)  #: emitted every time the hdf path changes

    sig_channel_doubleclicked: QtCore.pyqtSignal = QtCore.pyqtSignal(dict)  #: emitted when a ``Channel`` or ``Curve`` item is double clicked

    def __init__(self, parent, list_widgets: List[ListWidget]):
        """
        :param parent: parent dockWidget
        :param list_widgets: list of ListWidget ui objects, in the following order:
                             [sessions_list_widget, units_list_widget, channels_list_widget]
        """
        QtWidgets.QWidget.__init__(self, parent=parent)

        self.path: str = ''  #: system path to the hdf5 file
        self.file: h5py.File = None  #: h5py file handle

        self.session: str = ''  #: currently selected MSession
        self.sessions: List[str] = []  #: list of ``MSession`` options available in current file
        self.listw_sessions: ListWidget = list_widgets[0]  #: ui list of ``MSession`` options
        self.listw_sessions.itemClicked.connect(self.set_session)  #: sets the current ``MSession``
        self.listw_sessions.currentItemChanged.connect(self.set_session)

        self.unit: str = ''  #: currently selected ``MUnit``
        self.units: List[str] = []  #: list of ``MUnit`` options available in current ``MSession``
        self.listw_units: ListWidget = list_widgets[1]  #: ui list of ``MUnit`` options
        self.listw_units.itemClicked.connect(self.set_unit)  #: sets the current MUnit
        self.listw_units.currentItemChanged.connect(self.set_unit)

        self.channel: str = ''  #: currently selected ``Channel``
        self.channels: List[str] = []  #: list of ``Channel`` options available in current ``MUnit``
        self.listw_channels: ListWidget = list_widgets[2]  #: ui list of ``Channel`` options
        self.listw_channels.itemClicked.connect(self.set_channel)  #: sets the current ``Channel``
        self.listw_channels.currentItemChanged.connect(self.set_channel)
        self.listw_channels.itemDoubleClicked.connect(self._channel_clicked)

    def set_file_path(self, path: str):
        """
        set the path to the .mesc file

        :param path: path to the .mesc hdf5 file
        """
        f = h5py.File(path, mode='r')

        # Check if "MSession_X" keys exist in the file
        has_session_key = any(
            k.startswith('MSession') for k in f.keys()
        )

        if not has_session_key:
            raise TypeError(
                "The chosen file does not appear to be a valid "
                "`.mesc` file since it does not contain any "
                "'MSession_X' data group(s)."
            )

        self.path = path
        self.file = f

        # just clear out everything
        self.session: str = ''
        self.sessions: List[str] = list(self.file.keys())
        self._sort_list(self.sessions)
        self.listw_sessions.setItems(self.sessions)

        self.unit: str = ''
        self.units: List[str] = []
        self.listw_units.clear()

        self.channel: str = ''
        self.channels: List[str] = []
        self.listw_units.clear()

        self.sig_hpath_changed.emit(self.get_hpath(dict))

    def close_file(self):
        """
        Close the h5py file handle that is currently open.
        """
        if self.file is None:
            raise ValueError(
                'No file is currently open'
            )

        # close the hpy file handle
        self.file.close()

        # reset attributes to None
        self.file = None
        self.path = ''

        self.session: str = ''
        self.sessions: List[str] = []
        self.listw_sessions.clear()

        self.unit: str = ''
        self.units: List[str] = []
        self.listw_units.clear()

        self.channel: str = ''
        self.channels: List[str] = []
        self.listw_channels.clear()

        self.sig_hpath_changed.emit(self.get_hpath(dict))

    def set_session(self, key: Union[str, QtWidgets.QListWidgetItem]):
        """
        Set the MSession option

        :param key: a valid ``MSession``
        :return:
        """
        key = self._get_listw_text(key)
        if key is None:
            return

        if key not in self.sessions:
            raise KeyError(f"Session <{key}> not found in current file")

        self.session = key

        self.unit = ''
        self.units = list(self.file[self.session].keys())
        self._sort_list(self.units)
        self.listw_units.setItems(self.units)

        self.channel = ''
        self.channels: List[str] = []
        self.listw_channels.clear()

        self.sig_hpath_changed.emit(self.get_hpath(dict))

    def set_unit(self, key: Union[str, QtWidgets.QListWidgetItem]):
        """
        Set the MUnit option

        :param key: a valid ``MUnit``
        :return:
        """
        key = self._get_listw_text(key)
        if key is None:
            return

        if isinstance(key, QtWidgets.QListWidgetItem):
            key = key.text()

        if key not in self.units:
            raise KeyError(f"Unit <{key}> not found in current MSession")

        self.unit = key

        self.channel = ''
        self.channels = list(self.file[self.session][self.unit].keys())
        self._sort_list(self.channels)
        self.listw_channels.setItems(self.channels)
        self.listw_channels.setCurrentRow(0)

        self.sig_hpath_changed.emit(self.get_hpath(dict))

    def set_channel(self, key: Union[str, QtWidgets.QListWidgetItem]):
        """
        Set a Channel or Curve option

        :param key: a valid ``Channel`` or ``Curve``
        :return:
        """
        key = self._get_listw_text(key)
        if key is None:
            return

        if key not in self.channels:
            raise KeyError(f"Channel <{key}> not found in current MUnit")

        self.channel = key

        self.sig_hpath_changed.emit(self.get_hpath(dict))

    def _channel_clicked(self):
        if self.channel:
            self.sig_channel_doubleclicked.emit(self.get_hpath(dict))

    def get_hpath(self, astype: type) -> Union[str, list, dict]:
        """
        get the current hdf path

        :param astype: one of `str`, `list`, or `dict`
        :return: the hdf path as the chosen data type
        """
        path = [self.session, self.unit, self.channel]

        if astype is str:
            return '/'.join(
                filter(None, path)
            )

        elif astype is list:
            return path

        elif astype is dict:
            return {
                'session':  self.session,
                'unit':     self.unit,
                'channel': self.channel
            }

    def _get_listw_text(self, item: [str, QtWidgets.QListWidgetItem]) -> str:
        if isinstance(item, QtWidgets.QListWidgetItem):
            return item.text()
        else:
            return item

    def _sort_list(self, l: list):
        l.sort(
            key=lambda name: int(
                regex_sub(r"[^0-9]*", '', name)
            )
        )


class ModuleGUI(QtWidgets.QDockWidget):
    def __init__(self, parent, viewer_ref):  # when a ModuleGUI() is instantiated the
                                             # Viewer Window passes the viewer reference (viewer_ref)
                                             # which can be used to interface with the viewer
        self.vi = ViewerUtils(viewer_ref)  # this is used to interface with the viewer
        QtWidgets.QDockWidget.__init__(self, parent)

        self.ui = Ui_DockWidget()
        self.ui.setupUi(self)
        self.setWindowTitle('.mesc importer')

        self.ui.pushButton_open_file.clicked.connect(self.set_file)
        self.ui.pushButton_close_file.clicked.connect(self.close_file)
        self.ui.pushButton_import_image.clicked.connect(self.import_recording)

        # the list widgets for navigation
        list_widgets = [
            self.ui.listWidget_session,
            self.ui.listWidget_unit,
            self.ui.listWidget_channel
        ]

        self.mesc_navigator = MEScNavigator(self, list_widgets)  #: instance of ``MEScNavigator``
        self.mesc_navigator.sig_hpath_changed.connect(self._set_comment_line)
        self.mesc_navigator.sig_hpath_changed.connect(self._enable_import_button)

        self.mesc_navigator.sig_channel_doubleclicked.connect(
            self._channel_doubleclicked
        )

        self.plot_widgets = []  #: list of plot widgets

    @present_exceptions(
        'Could not load file',
        'The following error occurred while trying to load hte .mesc file.'
    )
    @use_open_file_dialog('Choose .mesc file', None, ['*.mesc'])
    def set_file(self, path: str, *args):
        """
        Create an h5py file handle from the `.mesc` file at the given ``path``.

        ``*args`` are not used, its just there for compatibility with the decorator.

        :param path: path to the `.mes` file
        :type path: str
        :param args: not used
        """
        self.mesc_navigator.set_file_path(path)

        # set the name of the current file in the GUI label
        self.ui.label_current_file_name.setText(
            os.path.basename(path)
        )

    @present_exceptions(
        'Could not close file',
        'The following error occurred while trying to close the file handle.'
    )
    def close_file(self, *args):
        """
        Close the file handle

        :param args: ``*args`` not used, just there for compatibility with the decorator
        """
        self.mesc_navigator.close_file()
        self.ui.label_current_file_name.clear()

    def _enable_import_button(self, hpath: dict):
        if hpath['channel']:
            self.ui.pushButton_import_image.setEnabled(True)
            self.ui.pushButton_import_stim_map.setEnabled(True)

        else:
            self.ui.pushButton_import_image.setEnabled(False)
            self.ui.pushButton_import_stim_map.setEnabled(False)

    def _channel_doubleclicked(self, hpath: dict):
        if hpath['channel'].startswith('Channel'):
            self.import_recording()

        elif hpath['channel'].startswith('Curve'):
            f = self.mesc_navigator.file
            hpath = self.mesc_navigator.get_hpath(str)

            ys = f[hpath]['CurveDataYRawData'][()]

            if 'CurveDataXRawData' in f[hpath].keys():
                xs = f[hpath]['CurveDataXRawData'][()]
            else:
                xs = np.arange(ys.size)

            pw = PlotWidget(parent=None)
            pw.plot(xs, ys, title=hpath)
            pw.show()

            self.plot_widgets.append(pw)

    def _set_comment_line(self, hpath: dict):
        if hpath['unit']:
            f = self.mesc_navigator.file
            comment = f[hpath['session']][hpath['unit']].attrs['Comment']
            comment = ascii_to_str(comment)
            self.ui.lineEdit_comment.setText(comment)

        else:
            self.ui.lineEdit_comment.clear()

    @present_exceptions(
        'Could not import recording',
        'The following error occurred while trying to import the chosen recording'
    )
    def import_recording(self, *args):
        """
        Imports the chosen recording into the Viewer Work Environment based
        on the user selected hpath from the list widgets

        ``*args`` not used, just there for compatibility with the decorator
        """
        if not self.vi.discard_workEnv():
            return

        # Load the image sequence stored in this `MUnit_X`
        f = self.mesc_navigator.file  # hdf file handle
        hpath = self.mesc_navigator.get_hpath(str)  # hdf path
        img_seq = bitwise_not(  # load & invert image array
            f[hpath][()]
        )

        sess = self.mesc_navigator.get_hpath(dict)['session']
        unit = self.mesc_navigator.get_hpath(dict)['unit']
        # get the time for a single frame
        frame_time = \
            f[sess][unit].attrs['ZAxisConversionConversionLinearScale'] + \
            f[sess][unit].attrs['ZAxisConversionConversionLinearOffset']

        z_units = ascii_to_str(f[sess][unit].attrs['ZAxisConversionUnitName'])
        if z_units != 'ms':
            raise TypeError(f"Z-axis units <{z_units}> not supported")

        # get the sampling rate
        fps = 1 / (frame_time / 1000)  # for now just assuming the frame time are always in milliseconds

        # get the date of the recording
        date = datetime.fromtimestamp(
            f[sess][unit].attrs['MeasurementDatePosix']
        )

        # format the date into YYYYMMDD_HHMMSS
        date_str = date.strftime('%Y%m%d_%H%M%S')

        # create a minimal meta data dict that is required for the Viewer Work Environment.
        # the keys ``fps``, ``origin`` and ``date`` are mandatory
        meta = \
            {
                'fps': fps,
                'origin': 'femtonics .mesc',
                'date': date_str,
            }

        # Create an ``ImgData`` object, see the ImgData class under Viewer Core API for more information
        imgdata = ImgData(img_seq.T, meta)

        # Create a new Viewer Work Environment object using this ``ImgData`` object
        self.vi.viewer.workEnv = ViewerWorkEnv(imgdata)

        # Update the GUI according to the new ViewerWorkEnv
        self.vi.update_workEnv()

        # set the "Current Image sequence" in the Viewer
        self.vi.viewer.ui.label_curr_img_seq_name.setText(hpath)



# -*- coding: utf-8 -*-
"""
@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

from ...common.qdialogs import *
from ...common.configuration import proj_cfg
from ..core.common import ViewerUtils
import numpy as np
from .pytemplates.caiman_importer_pytemplate import *
from caiman.utils.utils import load_dict_from_hdf5
import os


class ModuleGUI(QtWidgets.QDockWidget):
    def __init__(self, parent, viewer_reference):
        self.vi = ViewerUtils(viewer_reference)
        QtWidgets.QDockWidget.__init__(self, parent)

        self.ui = Ui_DockWidget()
        self.ui.setupUi(self)

        self.ui.pushButton_select_file.clicked.connect(self.select_file)
        self.ui.pushButton_import.clicked.connect(self.import_file)

        self.cnmf_data = None
        self.path = None

    def _load_data(self, path: str):
        try:
            self.cnmf_data = load_dict_from_hdf5(path)
        except Exception as e:
            self.cnmf_data = None
            self.path = None
            raise ValueError("Error while opening the file, make sure it is a valid Caiman HDF5 output file.\n")
        else:
            self.path = path

    @present_exceptions('Error importing caiman data')
    @use_open_file_dialog('Select HDF5 output file', exts=['*.hdf5', '*.h5', '*.hdf'])
    def select_file(self, path: str, *args):
        self._load_data(path)
        self.ui.label_file.setText(path)

    @present_exceptions('Error importing caiman data')
    def import_file(self, path: str = None):
        # Ask user if clear work env if it's not empty
        if hasattr(self.vi.viewer.workEnv, 'roi_manager'):
            if self.vi.viewer.workEnv.roi_manager is not None:
                if len(self.vi.viewer.workEnv.roi_manager.roi_list) > 0:
                    if QtWidgets.QMessageBox.warning(self, 'ROIs currently present in work environment',
                                                     'This will elar ROIs that are currently in the work '
                                                     'environment continue?',
                                                     QtWidgets.QMessageBox.Yes, QtWidgets.QMessageBox.No) == \
                            QtWidgets.QMessageBox.No:
                        return

        if (path is None) and (self.path is None):
            raise ValueError("Must specify path as an argument")

        # If this was used in command line mode
        elif path:
            self._load_data(path)

        self.vi.viewer.status_bar_label.showMessage('Importing CNMF ROIs from HDF5 file, please wait...')

        # Load data using the appropriate back-end manager
        if self.vi.viewer.workEnv.imgdata.ndim == 4:
            roi_manager_gui = self.vi.viewer.parent().get_module('roi_manager')
            roi_manager_gui.start_backend('VolCNMF')

            roi_manager_gui.manager.add_all_components(
                self.cnmf_data,
                input_params_dict=self.cnmf_data['params'],
            )
        else:
            roi_manager_gui = self.vi.viewer.parent().get_module('roi_manager')
            roi_manager_gui.start_backend('CNMFROI')

            roi_manager_gui.manager.add_all_components(
                self.cnmf_data,
                input_params_dict=self.cnmf_data['params'],
            )

        # TODO: ADD THE INPUT PARAMS FROM THE HDF5 FILE TO THE HISTORY TRACE, NEEDS TO OVERWRITE IF ALREADY EXISTS.

        self.vi.viewer.status_bar_label.showMessage('Finished CNMF ROIs!')



# -*- coding: utf-8 -*-
# from . import functions
import numpy as np
import pandas as pd
from ..base import BasePlotWidget
from ...utils import WidgetRegistry, get_colormap
from ....analysis import Transmission
from ....pyqtgraphCore.widgets.MatplotlibWidget import MatplotlibWidget
from ....pyqtgraphCore.graphicsItems.ScatterPlotItem import ScatterPlotItem
from ..datapoint_tracer import DatapointTracerWidget
from matplotlib.axes import Axes
from .control_widget import *
from ....pyqtgraphCore.widgets.ComboBox import ComboBox
from math import ceil
from itertools import product as iter_product
from tqdm import tqdm
from ....common.qdialogs import *
from uuid import UUID
from collections import OrderedDict


def tonumeric(s: str):
    try:
        if float(s).is_integer():
            return int(s)
        else:
            return float(s)
    except:
        return False


def get_tuning_curves(
        curve: np.ndarray,
        stim_maps: dict,
        method: str = 'mean',
        start_offset: int = 0,
        end_offset: int = 0,
        include_unlabelled: bool = False,
) -> pd.Series:
    """
    Returns a pandas series with tuning curves for all stimuli for a single curve

    :param curve: curve to create tuning curves from
    :param stim_maps: dict of stimulus maps
    :param method: stats method, such as "mean", "max", "median", etc.
    :param method_assign_stim: assign the "TUNE_<stim_name>" label according to min or max
    :param start_offset: start offset (in frames) for stimulus period extraction
    :param end_offset: end offset (in frames) for stimulus period extraction
    :return: pandas series:
                Tuning curves for each stimulus type are: "_TUNE_CURVE_<stim_name>" entries.
                "TUNE_<stim_name>" is assigned as the stimulus with the
                min or max value in the tuning curve

    """
    # empty output dict
    d = OrderedDict.fromkeys(
        [f"TUNE_CURVE_{k}_xlabels" for k in stim_maps.keys()] +   # used for the tuning curve
        [f"_TUNE_CURVE_{k}_yvals" for k in stim_maps.keys()] +
        [f"TUNE_MAX_{k}" for k in stim_maps.keys()] +  # used for stimulus at argmax(tuning_curve)
        [f"TUNE_MIN_{k}" for k in stim_maps.keys()]    # used for stimulus at argmin(tuning_curve)
    )

    # each stimulus datafame
    for stim_type in stim_maps.keys():
        stim_df = stim_maps[stim_type]  # get the stimulus dataframe

        # stimulus names will be stored as str with max length 32 char
        stim_array = np.empty(curve.size, dtype=np.dtype('<U32'))
        stim_array[:] = "None"

        # fill stim_array
        # stim_array is used for fancy indexing of `curve`
        # to extract the curve values during the various
        # stimuli
        for ix, r in stim_df.iterrows():  # iterate through the stimulus periods
            # apply any offsets
            start_ix = r['start'] + start_offset
            end_ix = r['end'] + end_offset

            # add this stimulus period to the stim_array
            stim_array[int(start_ix):int(end_ix)] = str(r['name'])  # store stim name as str

        # instantiate lists for the tuning curves
        all_stimuli = np.unique(stim_array)  # [stim_array != ''])
        if not include_unlabelled:
            all_stimuli = all_stimuli[all_stimuli != "None"]

        ln = [tonumeric(v) for v in all_stimuli if tonumeric(v) is not False]
        ls = [v for v in all_stimuli if tonumeric(v) is False]
        ln.sort()
        ls.sort()
        all_stimuli = np.array(ls + ln, dtype=np.dtype('<U32'))

        xs = np.empty(all_stimuli.size, dtype=np.dtype('<U32'))
        #         xs = []
        ys = np.empty(all_stimuli.size)

        # create the tuning curves
        for i, stimulus in enumerate(all_stimuli):
            xs[i] = stimulus  # stimulus name

            # get ALL the y values from the curve where
            # this specific stimulus is present
            y = curve[stim_array == stimulus]

            # get the mean, or max, etc. for the stimulus period
            ys[i] = getattr(np, method)(y)

        #         xs.append("None")
        #         ys.append(curve[np.isnan(stim_array)].mean())

        # the tuning curve
        d[f"TUNE_CURVE_{stim_type}_xlabels"] = xs
        d[f"_TUNE_CURVE_{stim_type}_yvals"] = ys

        # stimulus name at argmax() and argmin() of tuning curve
        d[f"TUNE_MAX_{stim_type}"] = xs[np.argmax(ys)]
        d[f"TUNE_MIN_{stim_type}"] = xs[np.argmin(ys)]

    return pd.Series(d)


class ControlDock(QtWidgets.QDockWidget):
    sig_sample_changed = QtCore.pyqtSignal(str)
    sig_roi_changed = QtCore.pyqtSignal(str)
    sig_params_changed = QtCore.pyqtSignal(dict)

    def __init__(self, parent):
        QtWidgets.QDockWidget.__init__(self, parent=parent)
        self.ui = Ui_ControlsDock()
        self.ui.setupUi(self)

        self.ui.comboBox_method.setItems(
            ['mean', 'median', 'max', 'min']
        )

        self.widget_registry = WidgetRegistry()

        self.widget_registry.register(
            self.ui.comboBox_data_column,
            setter=self.ui.comboBox_data_column.setText,
            getter=self.ui.comboBox_data_column.currentText,
            name='data_column'
        )

        self.widget_registry.register(
            self.ui.comboBox_method,
            setter=self.ui.comboBox_method.setText,
            getter=self.ui.comboBox_method.currentText,
            name='method'
        )

        self.widget_registry.register(
            self.ui.spinBox_start_offset,
            setter=self.ui.spinBox_start_offset.setValue,
            getter=self.ui.spinBox_start_offset.value,
            name='start_offset',
        )

        self.widget_registry.register(
            self.ui.spinBox_end_offset,
            setter=self.ui.spinBox_end_offset.setValue,
            getter=self.ui.spinBox_end_offset.value,
            name='end_offset',
        )

        self.widget_registry.register(
            self.ui.comboBox_DPT_column,
            setter=self.ui.comboBox_DPT_column.setText,
            getter=self.ui.comboBox_DPT_column.currentText,
            name='dpt_column'
        )

        self.widget_registry.register(
            self.ui.checkBoxIncludeUnlabelled,
            setter=self.ui.checkBoxIncludeUnlabelled.setChecked,
            getter=self.ui.checkBoxIncludeUnlabelled.isChecked,
            name='include_unlabelled'
        )

        self.ui.pushButton_set.clicked.connect(self._emit_data)

        self.ui.listWidget_samples.currentTextChanged.connect(self.sig_sample_changed.emit)
        self.ui.listWidget_rois.currentTextChanged.connect(self.sig_roi_changed.emit)

    def _emit_data(self):
        self.sig_params_changed.emit(
            self.widget_registry.get_state()
        )

    def fill_widget(self, samples: list, data_columns: list):
        self.ui.comboBox_data_column.setItems(data_columns)
        self.ui.listWidget_samples.setItems(samples)
        self.ui.comboBox_DPT_column.setItems(data_columns)

    def get_state(self) -> dict:
        return self.widget_registry.get_state()

    def set_state(self, state: dict):
        self.widget_registry.set_state(state)


class PlotArea(MatplotlibWidget):
    def __init__(self, parent):
        MatplotlibWidget.__init__(self)
        self.axs: Axes = None  #: array of axis objects used for drawing the means plots, shape is [nrows, ncols]
        self.setParent(parent)
        self.ncols = 2
        self.nrows: int = None  # determined at plot time

    def set_plots(
            self,
            tuning_curves: Dict[str, np.ndarray],
            y_units: List[str]
    ):
        """
        Set the subplots

        :param input_arrays: padded input arrays (2D),  shape is [num_samples, padded_peak_curve_length]
        :param n_clusters: number of clusters
        :param y_pred: cluster predictions (labels)
        :param xzero_pos: set the zero position as the 'zero' position of the input array or the 'maxima' of the input array
        :param error_band: Type of error band to show, one of either 'ci' or 'std'
        """
        self.clear()

        stim_types = list(tuning_curves.keys())

        if len(stim_types) == 1:
            self.ncols = 1
        elif len(stim_types) < 5:
            self.ncols = 2
        else:
            self.ncols = 3

        self.nrows = ceil(len(tuning_curves.keys()) / self.ncols)

        self.axs = self.fig.subplots(self.nrows, self.ncols)
        self.fig.tight_layout()

        for i, plot_ix in enumerate(
                iter_product(range(self.nrows), range(self.ncols))
        ):
            if not i < len(stim_types):
                break

            stim_type = stim_types[i]
            data = tuning_curves[stim_type]

            if len(stim_types) == 1:
                plot = self.axs
            else:
                plot = self.axs[plot_ix]

            plot.plot(data[0], data[1], c='k')
            plot.set_xlabel(stim_type)
            plot.set_ylabel(f"{y_units} response")

        self.draw()

    def clear(self):
        self.fig.clear()


class TuningCurvesWidget(QtWidgets.QMainWindow, BasePlotWidget):
    drop_opts = []
    sig_output_changed = QtCore.pyqtSignal(Transmission)  #: Emits output Transmission containing tuning curves data

    def __init__(self):
        QtWidgets.QMainWindow.__init__(self, parent=None)
        BasePlotWidget.__init__(self)

        self.setWindowTitle('Tuning Curve Plots')

        self.plot = PlotArea(parent=self)
        self.setCentralWidget(self.plot)

        self.control_widget = ControlDock(self)
        self.addDockWidget(
            QtCore.Qt.LeftDockWidgetArea, self.control_widget
        )

        self.control_widget.sig_params_changed.connect(self.update_tuning_curves)

        self.sample_id = None
        self.roi_uuid = None

        self.roi_uuid_map: dict = None

        self.control_widget.sig_sample_changed.connect(self.set_rois_widget)
        self.control_widget.sig_roi_changed.connect(self.update_plot)

        self.control_widget.ui.pushButton_save.clicked.connect(self.save_plot_dialog)

        self.datapoint_tracer = DatapointTracerWidget()

        self.datapoint_tracer_dockwidget = QtWidgets.QDockWidget(parent=self)
        self.datapoint_tracer_dockwidget.setWidget(self.datapoint_tracer)
        self.addDockWidget(
            QtCore.Qt.RightDockWidgetArea,
            self.datapoint_tracer_dockwidget
        )

        # combobox so user can choose which stims to display using colored regions

        # create hlayout
        hlayout_datapoint_tracer_stimulus_options = QtWidgets.QHBoxLayout()
        # qlabel
        datapoint_tracer_stimulus_options_qlabel = QtWidgets.QLabel(
            self.datapoint_tracer.ui.groupBoxImageViewAndCurve
        )
        datapoint_tracer_stimulus_options_qlabel.setText(
            'Illustrate stimulus type: '
        )

        # combobox
        self.datapoint_tracer_stimulus_options_combobox = ComboBox(
            self.datapoint_tracer.ui.groupBoxImageViewAndCurve
        )
        # add them to the hlayout
        hlayout_datapoint_tracer_stimulus_options.addWidget(
            datapoint_tracer_stimulus_options_qlabel
        )
        hlayout_datapoint_tracer_stimulus_options.addWidget(
            self.datapoint_tracer_stimulus_options_combobox
        )
        self.datapoint_tracer_stimulus_options_combobox.currentTextChanged.connect(
            self.set_datapoint_tracer_stimulus_regions
        )
        # add hlayout under the curve plot
        self.datapoint_tracer.ui.verticalLayout_groupBoxImageViewAndCurve.addLayout(
            hlayout_datapoint_tracer_stimulus_options
        )

        # datapoint tracer curve plot legend for stimulus region colors
        self.legend = self.datapoint_tracer.ui.graphicsViewPlot.plotItem.addLegend()
        self.legend.setParentItem(
            self.datapoint_tracer.ui.graphicsViewPlot.getPlotItem()
        )
        self.color_legend_items = []
        self.pseudo_plots = []

        self.current_stim_region_plot: str = None

        self.update_live = True

    def update_tuning_curves(self, params: dict):
        data_column = params['data_column']
        method = params['method']
        start_offset = params['start_offset']
        end_offset = params['end_offset']

        tqdm().pandas()

        self.transmission.df[
                    [f"TUNE_CURVE_{s}_xlabels" for s in self.transmission.STIM_DEFS] +
                    [f"_TUNE_CURVE_{s}_yvals" for s in self.transmission.STIM_DEFS] +
                    [f"TUNE_MAX_{s}" for s in self.transmission.STIM_DEFS] +
                    [f"TUNE_MIN_{s}" for s in self.transmission.STIM_DEFS]
                 ] = \
            self.transmission.df.progress_apply(
                lambda r: get_tuning_curves(
                    curve=r[data_column],
                    stim_maps=r['stim_maps'][0][0],
                    method=method,
                    start_offset=start_offset,
                    end_offset=end_offset
                ), axis=1
            )

        self.send_output_transmission()

    @BasePlotWidget.signal_blocker
    def set_input(self, transmission: Transmission):
        """Set the input transmission"""
        if (self._transmission is None) or self.update_live:
            super(TuningCurvesWidget, self).set_input(transmission)
            self.plot.clear()

            # set the stimulus options for visualizing them as colored regions
            self.datapoint_tracer_stimulus_options_combobox.setItems(
                self.transmission.STIM_DEFS
            )

    @BasePlotWidget.signal_blocker
    def set_rois_widget(self, sample_id):
        self.plot.clear()

        self.sample_id = sample_id
        roi_uuids = \
            self.transmission.df[
                self.transmission.df['SampleID'] == self.sample_id
            ].uuid_curve.unique().tolist()

        # user friendly integer map to the ROIs
        self.roi_uuid_map = dict(
            zip(
                map(str, range(len(roi_uuids))),
                roi_uuids
            )
        )

        self.control_widget.ui.listWidget_rois.clear()
        self.control_widget.ui.listWidget_rois.setItems(self.roi_uuid_map.keys())
        self.current_stim_region_plot = None

    @BasePlotWidget.signal_blocker
    def fill_control_widget(self, data_columns: list, categorical_columns: list, uuid_columns: list):
        samples = list(self.transmission.df['SampleID'].unique())
        self.control_widget.fill_widget(samples=samples, data_columns=data_columns)

    @present_exceptions(
        'Plot error',
        'Plot error. Make sure you have selected appropriate data columns and parameters'
    )
    def update_plot(self, *args, **kwargs):
        if self.control_widget.ui.listWidget_rois.currentItem() is None:
            return
        
        roi_ix = self.control_widget.ui.listWidget_rois.currentItem().text()
        uuid_curve = self.roi_uuid_map[roi_ix]

        r = self.transmission.df[
            self.transmission.df['uuid_curve'] == uuid_curve
        ]

        # get the tuning curves for all stims
        # pass it as a x-y list
        tuning_curves = \
            {
                s: [
                    r[f"TUNE_CURVE_{s}_xlabels"].item(),
                    r[f"_TUNE_CURVE_{s}_yvals"].item()
                ] for s in self.transmission.STIM_DEFS
            }

        # curve made using mean response, or max response etc.
        y_units = self.get_plot_opts()['method']

        self.plot.set_plots(tuning_curves, y_units)

        block_id = r._BLOCK_
        if isinstance(block_id, pd.Series):
            block_id = block_id.item()
        h = self.transmission.history_trace.get_data_block_history(block_id)

        dpt_column = self.control_widget.ui.comboBox_DPT_column.currentText()

        # self.datapoint_tracer.peak_region.clear_all()

        # TODO: Use the TimelineLinearRegion.add_linear_region to add
        #       stimulus period illustrations for all stimuli types
        #       Have some way to get a legend for the colors

        self.datapoint_tracer.set_widget(
            datapoint_uuid=UUID(uuid_curve),
            data_column_curve=dpt_column,
            row=r,
            proj_path=self.transmission.get_proj_path(),
            history_trace=h,
            clear_linear_regions=False
        )

        self.set_datapoint_tracer_stimulus_regions()

    def set_datapoint_tracer_stimulus_regions(self):
        selected_stim_type = self.datapoint_tracer_stimulus_options_combobox.currentText()

        # don't redraw
        if self.current_stim_region_plot == selected_stim_type:
            return

        self.current_stim_region_plot = selected_stim_type

        roi_ix = self.control_widget.ui.listWidget_rois.currentItem().text()
        uuid_curve = self.roi_uuid_map[roi_ix]

        r = self.transmission.df[
            self.transmission.df['uuid_curve'] == uuid_curve
            ]

        stim_df = r['stim_maps'].item()[0][0][selected_stim_type].sort_values(by='start')

        stims = stim_df.name.unique().astype(str)

        if stims.size < 10:
            _cm = 'tab10'
        elif stims.size > 20:
            _cm = 'hsv'
        else:
            _cm = 'tab20'

        cmap = get_colormap(stims, _cm, output='pyqt')

        self.datapoint_tracer.peak_region.clear_all()

        region_samples = dict()

        for ix, stim_period in stim_df.iterrows():
            start = stim_period['start']
            end = stim_period['end']
            color = cmap[stim_period['name']]

            region = self.datapoint_tracer.peak_region.add_linear_region(
                start, end, color
            )

            if stim_period['name'] not in region_samples.keys():
                region_samples[stim_period['name']] = region

        self.clear_legend()

        for k in cmap.keys():
            p = ScatterPlotItem()
            p.setData(x=[0], y=[0], brush=cmap[k], symbol='s')
            self.color_legend_items.append(k)
            self.legend.addItem(p, k)
            self.pseudo_plots.append(p)

    def clear_legend(self):
        """Clear the legend"""
        for i in range(len(self.pseudo_plots)):
            del self.pseudo_plots[0]

        for name in self.color_legend_items:
            self.legend.removeItem(name)
        self.color_legend_items.clear()

    def send_output_transmission(self):
        """Send output Transmission containing cluster labels"""
        params = self.get_plot_opts()

        t = self.transmission.copy()
        t.history_trace.add_operation('all', operation='tuning_curves', parameters=params)

        self.sig_output_changed.emit(t)

    def save_plot(self, path):
        t = self.transmission.copy()

        params = self.get_plot_opts()

        self.transmission.history_trace.add_operation('all', operation='tuning_curves', parameters=params)
        super(TuningCurvesWidget, self).save_plot(path)

        self.transmission = t

    def set_update_live(self, b: bool):
        pass

    def get_plot_opts(self, drop: bool = False) -> dict:
        return self.control_widget.get_state()

    @BasePlotWidget.signal_blocker
    def set_plot_opts(self, opts: dict):
        # it will try to set the roi index without
        # any Sample having been selected
        try:
            self.control_widget.set_state(opts)
        except AttributeError:
            pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on April 19 2018

@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

from ..core.common import ViewerUtils
from ..core.viewer_work_environment import ViewerWorkEnv
from .pytemplates.mesfile_io_pytemplate import *
from .stimulus_mapping_mesfile import *
import traceback
import pandas as pd
from .stimulus_mapping import ModuleGUI as StimMapModuleGUI


class ModuleGUI(QtWidgets.QDockWidget):
    def __init__(self, parent, viewer_reference):
        self.vi = ViewerUtils(viewer_reference)
        self.parent = parent

        QtWidgets.QDockWidget.__init__(self, parent)
        self.ui = Ui_DockWidget()
        self.ui.setupUi(self)

        self.ui.btnOpenMesFile.clicked.connect(self.load_mesfile)
        self.ui.listwMesfile.itemDoubleClicked.connect(lambda sel: self.load_mesfile_selection(sel))
        self.ui.btnStimMapGUI.clicked.connect(self.open_stim_map_gui)

        self.stim_map_gui = None
        self.mesfile = None

    def load_mesfile(self):
        if not self.vi.discard_workEnv():
            return

        filelist = QtWidgets.QFileDialog.getOpenFileNames(self, 'Choose ONE mes file', '.', '(*.mes)')

        if len(filelist) == 0:
            return

        try:
            # Creates an instance of MES, see MesmerizeCore.FileInput
            self.mesfile = ViewerWorkEnv.load_mesfile(filelist[0][0])
            self.ui.listwMesfile.setEnabled(True)
            self.ui.listwMesfile.clear()
            # Get the references of the images, their descriptions, and add them to the list
            for i in self.mesfile.get_image_references():
                j = self.mesfile.image_descriptions[i]
                self.ui.listwMesfile.addItem(i + ': ' + j)

            # If Auxiliary voltage info is found in the mes file, ask the user if they want to map these to stimuli
            if len(self.mesfile.voltages_lists_dict) > 0:
                self.stim_map_gui = MesStimmapGUI(parent=None, viewer=self.vi.viewer)
                self.set_stim_map_gui()
                self.ui.btnStimMapGUI.setEnabled(True)
                if QtWidgets.QMessageBox.question(self, '', 'This .mes file contains auxilliary output voltage '
                                                            'information, would you like create Stimulus Maps now?',
                                                  QtWidgets.QMessageBox.Yes,
                                                  QtWidgets.QMessageBox.No) == QtWidgets.QMessageBox.Yes:
                    self.open_stim_map_gui()
            else:
                self.ui.btnStimMapGUI.setDisabled(True)

        except (IOError, IndexError) as e:
            QtWidgets.QMessageBox.warning(self, 'IOError or IndexError',
                                          "There is an problem with the files you've selected:\n" + traceback.format_exc(),
                                          QtWidgets.QMessageBox.Ok)
        return

    def load_mesfile_selection(self, s: QtWidgets.QListWidgetItem):
        if not self.vi.discard_workEnv():
            return
        try:
            self.vi.viewer.status_bar_label.showMessage('Loading image sequence from mesfile please wait...')
            self.vi.viewer.workEnv = ViewerWorkEnv.from_mesfile(self.mesfile, s.text().split(': ')[0])
        except KeyError as ke:
            QtWidgets.QMessageBox.warning(self, f'KeyError: {ke}', traceback.format_exc(), QtWidgets.QMessageBox.Ok)
            self.vi.viewer.status_bar_label.clearMessage()
            return

        except Exception:
            QtWidgets.QMessageBox.warning(self, 'Error', 'Error opening the selected ' + \
                                          'image in the currently open mes file.\n' +
                                          traceback.format_exc(), QtWidgets.QMessageBox.Ok)
            self.vi.viewer.status_bar_label.clearMessage()
            return
        self.vi.viewer.ui.label_curr_img_seq_name.setText(s.text())
        self.vi.viewer.status_bar_label.showMessage('Setting Stimulus Map...')
        self.set_stimulus_map()
        self.vi.viewer.status_bar_label.showMessage('Finished setting stimulus map!')
        self.vi.update_workEnv()
        self.vi.enable_ui(True)

    def open_stim_map_gui(self):
        self.stim_map_gui.show()

    def set_stim_map_gui(self):
        for channel in self.mesfile.voltages_lists_dict.keys():
            self.stim_map_gui.add_stim_type(channel)
            for voltage in self.mesfile.voltages_lists_dict[channel]:
                pd_series = pd.Series(data={'voltage': '%.1f' % voltage,
                                            'name': '',
                                            'color': '#FFFFFF'
                                            })
                self.stim_map_gui.tabs[channel].add_row(pd_series)

    def set_stimulus_map(self):
        if self.stim_map_gui.voltage_mappings is None:
            QtWidgets.QMessageBox.information(self, 'Voltage mappings not set',
                                              'You have not set the mesfile voltage -> stimulus mappings')
            return

        smm = self.parent.run_module(StimMapModuleGUI)
        assert isinstance(smm, StimMapModuleGUI)

        stimulus_dataframes = {}
        meta = self.vi.viewer.workEnv.meta['orig_meta']

        voltage_mappings = self.stim_map_gui.voltage_mappings

        for stim_type in voltage_mappings.keys():
            channel = voltage_mappings[stim_type]['channel']

            try:
                y = meta[channel]['y']
                x = meta[channel]['x'][1]

                firstFrameStartTime = meta['FoldedFrameInfo']['firstFrameStartTime']
                frameTimeLength = meta['FoldedFrameInfo']['frameTimeLength']

                current_map = []

                for i in range(0, y.shape[1] - 1):
                    voltage = '%.1f' % y[1][i]

                    tstart_frame = int(((y[0][i] * x) - firstFrameStartTime) / frameTimeLength)

                    if tstart_frame < 0:
                        tstart_frame = 0

                    tend_frame = int(((y[0][i + 1] * x) - firstFrameStartTime) / frameTimeLength)

                    #v = voltage# + ' V: '

                    mapping = voltage_mappings[stim_type]['dataframe'][
                        voltage_mappings[stim_type]['dataframe']['voltage'] == voltage]
                    try:
                        name = mapping['name'].item()
                    except:
                        name = mapping['name']
                    try:
                        color = mapping['color'].item()
                    except:
                        color = mapping['color']

                    current_map.append({'name': name,
                                        'start': tstart_frame,
                                        'end': tend_frame,
                                        'color': color
                                        })
                # stimulus_dataframes[stim_type] = {'units': 'frames', 'dataframe': pd.DataFrame(current_map)}
                stimulus_dataframes[stim_type] = pd.DataFrame(current_map)

            except (KeyError, IndexError):
                QtWidgets.QMessageBox.information(None, 'FYI: Missing channels in current image',
                                                  'Voltage values not found for stimulus type: "' + stim_type + \
                                                  '" in channel <' + channel + '>.\n' + traceback.format_exc())

        smm.set_all_data(stimulus_dataframes)
        print(stimulus_dataframes)
        smm.export_to_work_env()
        try:
            smm.ui.comboBoxShowTimelineChoice.setCurrentIndex(1)
        except:
            pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on June 5 2018

@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

from ..common import configuration, system_config_window, doc_pages, get_project_manager, set_project_manager
from ..pyqtgraphCore.console import ConsoleWidget
from .pytemplates.welcome_window_pytemplate import *
# from ..project_manager import ProjectManager
from ..common import get_window_manager, is_mesmerize_project, get_session_id
# from viewer.modules import batch_manager
import os
from ..common import start
from ..viewer.modules.batch_manager import ModuleGUI as BatchModuleGUI
from glob import glob
from ..plotting import open_plot_file
from functools import partial
from .qdialogs import *
import re
from .. import __version__


class MainWindow(QtWidgets.QMainWindow):
    def __init__(self):
        QtWidgets.QMainWindow.__init__(self)
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)

        self.setWindowTitle('Mesmerize - Welcome Window')

        self.window_manager = get_window_manager()

        self.ui.btnProjectBrowser.setVisible(False)
        self.ui.labelProjectBrowser.setVisible(False)

        mpath = os.path.abspath(__file__)
        mdir = os.path.dirname(mpath)

        self.ui.btnProjectBrowser.setIcon(QtGui.QIcon(mdir + '/icons/noun_917603_cc.png'))
        self.ui.btnProjectBrowser.setIconSize(QtCore.QSize(100, 100))
        self.ui.btnProjectBrowser.clicked.connect(self.show_project_browser)

        self.ui.btnNewProject.setIcon(QtGui.QIcon(mdir + '/icons/noun_1327089_cc.png'))
        self.ui.btnNewProject.setIconSize(QtCore.QSize(100, 100))
        self.ui.btnNewProject.clicked.connect(self.create_new_project)

        self.ui.btnOpenProject.setIcon(QtGui.QIcon(mdir + '/icons/noun_1327109_cc.png'))
        self.ui.btnOpenProject.setIconSize(QtCore.QSize(100, 100))
        self.ui.btnOpenProject.clicked.connect(self.open_project_dialog)

        sys_cfg = configuration.get_sys_config()
        self.ui.listWidgetRecentProjects.addItems(sys_cfg['recent_projects'])
        self.ui.listWidgetRecentProjects.itemDoubleClicked.connect(lambda item: self.open_project(item.text()))

        self.ui.btnViewer.setIcon(QtGui.QIcon(mdir + '/icons/noun_38902_cc.png'))
        self.ui.btnViewer.setIconSize(QtCore.QSize(24, 24))
        self.ui.btnViewer.setIconSize(QtCore.QSize(100, 100))
        self.ui.btnViewer.clicked.connect(self.open_new_viewer)

        self.ui.btnFlowchart.setIcon(QtGui.QIcon(mdir + '/icons/noun_907242_cc.png'))
        self.ui.btnFlowchart.setIconSize(QtCore.QSize(100, 100))
        self.ui.btnFlowchart.clicked.connect(self.open_new_flowchart)

        self.ui.actionDocs_homepage.triggered.connect(doc_pages['home'])
        self.ui.actionNew_Project_Docs.triggered.connect(doc_pages['new_project'])
        self.ui.actionSystem_Configuration_Docs.triggered.connect(doc_pages['sys_config'])

        self.ui.actionReport_issue_bug.triggered.connect(doc_pages['issue-tracker'])
        self.ui.actionQuestions_discussion.triggered.connect(doc_pages['gitter'])

        self.sys_config_gui = system_config_window.SystemConfigGUI()
        self.ui.actionSystem_Configuration.triggered.connect(self.sys_config_gui.show)

        self.initialize_console_widget()

        self.resize(914, 900)
        self.setWindowIcon(QtGui.QIcon(mdir + '/icons/main_icon.gif'))

        self._batch_manager = None

        self.ui.treeViewFlowcharts.setVisible(False)
        self.ui.labelProjectPlots.setVisible(False)
        self.ui.treeViewProjectPlots.setVisible(False)

        self.plot_windows = []

        self.flowcharts_dir = None
        self.plots_dir = None

        if not configuration.HAS_CAIMAN:
            QtWidgets.QMessageBox.information(
                self,
                'Caiman not found',
                'The caiman package could not be found in your python environment. '
                'Caiman features will not work.'
            )

        self.ui.action_version.setText(
            f"Version: {__version__}"
        )

    def initialize_console_widget(self):
        ns = {'configuration': configuration,
              'get_window_manager': get_window_manager,
              'this': self,
              'get_session_id': get_session_id
              }

        txt = '\n'.join(["Namespaces:",
                         "self as this",
                         "callables:",
                         "get_window_manager()",
                         "get_session_id()"])

        cmd_history_file = os.path.join(configuration.console_history_path, 'welcome_window.pik')
        console = ConsoleWidget(namespace=ns, text=txt, historyFile=cmd_history_file)

        self.ui.centralwidget.layout().addWidget(console)
        # self.ui.centralwidget.layout().addWidget(spacer)
        # self.ui.dockConsole.setWidget(console)

        # self.resizeDocks([self.ui.dockConsole], [235], QtCore.Qt.Vertical)
        # console.resize(self.ui.dockConsole.width(), 100)
        # self.ui.dockConsole.resize(self.ui.dockConsole.width(), 150)
        # console.input.resize(console.input.width(), 80)

        # self.ui.dockConsole.hide()

    def append_recent_projects_list(self):
        sys_cfg = configuration.get_sys_config()
        if self.project_manager.root_dir in sys_cfg['recent_projects']:
            sys_cfg['recent_projects'].remove(self.project_manager.root_dir)
        sys_cfg['recent_projects'].insert(0, self.project_manager.root_dir)
        if len(sys_cfg['recent_projects']) > 10:
            sys_cfg['recent_projects'] = sys_cfg['recent_projects'][:10]

        configuration.save_sys_config(sys_cfg)

    @use_open_dir_dialog('Choose location for a new project')
    @present_exceptions('Could not create new project', 'The following error occurred while creating a new project')
    def create_new_project(self, path, *args):
        # path = QtWidgets.QFileDialog.getExistingDirectory(self, 'Choose location for a new project')
        # if path == '':
        #     return

        name, start = QtWidgets.QInputDialog.getText(self, '', 'Project Name:', QtWidgets.QLineEdit.Normal, '')

        if not re.match("^[A-Za-z0-9_-]*$", name):
            QtWidgets.QMessageBox.warning(self, 'Invalid Name', 'Project Name can only contain alphanumeric characters')
            return

        if start and name != '':
            # try:
            proj_dir = os.path.join(path, name)
            self.project_manager = get_project_manager()
            self.project_manager.set(proj_dir)
            self.project_manager.setup_new_project()
            # except Exception:
            #     QtWidgets.QMessageBox.warning(self, 'Error!',
            #                                   'Could not create a new project.\n' +
            #                                   traceback.format_exc())
            #     return
        else:
            return
        self.ui.actionProject_Configuration.setEnabled(True)
        self.ui.actionProject_Configuration.triggered.connect(self.project_manager.show_config_window)

        self.set_proj_buttons_visible(False)
        self.append_recent_projects_list()

        self.setWindowTitle(
            f'{self.windowTitle()}'
            f' - '
            f'{os.path.basename(self.project_manager.root_dir)}'
        )

        # self.initialize_project_browser()

    @use_open_dir_dialog('Select Project Directory', '')
    def open_project_dialog(self, path, *args):
        self.open_project(path)

    @present_exceptions('Cannot open project', 'The selected directory is probably not a valid Mesmerize Project.')
    def open_project(self, path: str):
        is_mesmerize_project(path)
        self.project_manager = get_project_manager()
        self.project_manager.set(project_root_dir=path)

        self.project_manager.open_project()

        self.ui.actionProject_Configuration.setEnabled(True)
        self.ui.actionProject_Configuration.triggered.connect(self.project_manager.show_config_window)

        self.set_proj_buttons_visible(False)

        pb = start.project_browser()
        pb.reload_all_tabs()
        self.append_recent_projects_list()

        self.setWindowTitle(
            f'{self.windowTitle()}'
            f' - '
            f'{os.path.basename(self.project_manager.root_dir)}'
        )

    def show_project_browser(self):
        get_window_manager().project_browser.show()

    def set_proj_buttons_visible(self, b: bool):
        self.ui.btnNewProject.setVisible(b)
        self.ui.labelNewProj.setVisible(b)
        self.ui.btnOpenProject.setVisible(b)
        self.ui.labelOpenProject.setVisible(b)

        # self.ui.listWidgetRecentProjects.setVisible(b)
        # self.ui.labelRecentProjects.setVisible(b)

        self.ui.btnProjectBrowser.setHidden(b)
        self.ui.labelProjectBrowser.setHidden(b)

        self.set_flowcharts_list()
        self.set_plots_list()

    def set_flowcharts_list(self):
        self.ui.labelRecentProjects.setText('Project flowcharts')
        self.ui.listWidgetRecentProjects.setVisible(False)
        self.ui.treeViewFlowcharts.setVisible(True)

        self.flowcharts_dir = os.path.join(self.project_manager.root_dir, 'flowcharts')
        model = QtWidgets.QFileSystemModel()
        model.setRootPath(self.flowcharts_dir)
        model.setFilter(QtCore.QDir.NoDotAndDotDot | QtCore.QDir.Files | QtCore.QDir.Dirs)
        self.ui.treeViewFlowcharts.setModel(model)
        self.ui.treeViewFlowcharts.setRootIndex(model.index(self.flowcharts_dir))
        self.ui.treeViewFlowcharts.setColumnHidden(1, True)
        self.ui.treeViewFlowcharts.setColumnHidden(2, True)
        self.ui.treeViewFlowcharts.header().setSectionResizeMode(0, QtWidgets.QHeaderView.ResizeToContents)

        self.ui.treeViewFlowcharts.doubleClicked.connect(lambda m_ix: self.open_new_flowchart(model.filePath(m_ix)))

    def set_plots_list(self):
        self.ui.labelProjectPlots.setVisible(True)
        self.ui.treeViewProjectPlots.setVisible(True)

        self.plots_dir = os.path.join(self.project_manager.root_dir, 'plots')

        model = QtWidgets.QFileSystemModel()
        model.setRootPath(self.plots_dir)
        model.setFilter(QtCore.QDir.NoDotAndDotDot | QtCore.QDir.Files | QtCore.QDir.Dirs)
        self.ui.treeViewProjectPlots.setModel(model)
        self.ui.treeViewProjectPlots.setRootIndex(model.index(self.plots_dir))
        self.ui.treeViewProjectPlots.setColumnHidden(1, True)
        self.ui.treeViewProjectPlots.setColumnHidden(2, True)
        self.ui.treeViewProjectPlots.header().setSectionResizeMode(0, QtWidgets.QHeaderView.ResizeToContents)

        self.ui.treeViewProjectPlots.doubleClicked.connect(lambda m_ix: self.open_plot(model.filePath(m_ix)))

    @present_exceptions('Could not open plot', 'The selected file is probably not a valid plot file.\n'
                                               'The following error occurred:\n')
    def open_plot(self, filename: str):
        if not os.path.isfile(filename):
            return
        plot = open_plot_file(filename)
        self.plot_windows.append(plot)
        plot.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        plot.destroyed.connect(partial(self.plot_windows.remove, plot))
        plot.show()

    def open_new_viewer(self):
        w = get_window_manager().get_new_viewer_window()
        w.show()
        # start.viewer()

    def open_new_flowchart(self, filename: str = None):
        if filename is not None and not isinstance(filename, str):
            filename = None
        else:
            if not os.path.isfile(filename):
                return

        w = get_window_manager().get_new_flowchart(filename, parent=self)
        w.show()

    def get_batch_manager(self, run_batch: list = None, testing=False) -> BatchModuleGUI:
        if run_batch is not None:
            self._batch_manager = BatchModuleGUI(parent=None, run_batch=run_batch)
            self._batch_manager.show()
            return self._batch_manager

        elif self._batch_manager is None:
            if not testing:
                 QtWidgets.QMessageBox.information(None, 'No batch manager open', 'Choose a location for a new batch or create a new batch')

            self._batch_manager = BatchModuleGUI(parent=None, testing=testing)
            # self._batch_manager.show()
            return self._batch_manager
        else:
            return self._batch_manager
            # self._batch_manager.show()
            # self._batch_manager.show()

    def closeEvent(self, QCloseEvent):
        if QtWidgets.QMessageBox.warning(self, 'Close Application?',
                                         'Are you sure you want to close all Mesmerize windows?',
                                         QtWidgets.QMessageBox.Yes,
                                         QtWidgets.QMessageBox.No) == QtWidgets.QMessageBox.No:
            QCloseEvent.ignore()
        else:
            for w in QtWidgets.QApplication.topLevelWidgets():
                QCloseEvent.accept()
                w.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

from . import get_sys_config
import os
from stat import S_IEXEC
from time import time
from datetime import datetime


def make_workdir(suffix: str = '') -> str:
    date = datetime.fromtimestamp(time())
    time_str = date.strftime('%Y%m%d') + '_' + date.strftime('%H%M%S')
    workdir = '/work/' + os.environ['USER'] + '/' + time_str + '_' + suffix
    os.makedirs(workdir)
    return workdir


def make_runfile(module_path: str, workdir: str, args_str: str = None, filename: str = None,
                 pre_run: str = '', post_run: str = '') -> str:
    """
    :param module_path: absolute module path
    :param args_str:    str of args that is directly passed with the python command in the bash script
    :param workdir:     working directory
    :param filename:    optional, specific filename for the script
    :param pre_run:     optional, str to run before module is ran
    :param post_run:    optional, str to run after module has run

    :return: path to the shell script that can be run
    """
    if filename is None:
        sh_file = workdir + '/' + 'run.sh'
    else:
        sh_file = workdir + '/' + filename

    sys_cfg = get_sys_config()

    if sys_cfg['PATHS']['env_type'] == 'anaconda':
        env_path = sys_cfg['PATHS']['env']
        anaconda_dir = os.path.dirname(os.path.dirname(os.path.dirname(env_path)))
        env_name = os.path.basename(os.path.normpath(env_path))
        env_activation = 'export PATH=' + anaconda_dir + ':$PATH\nsource activate ' + env_name
    elif sys_cfg['PATHS']['env_type'] == 'virtual':
        env_path = sys_cfg['PATHS']['env']
        env_activation = 'source ' + env_path + '/bin/activate'
    else:
        raise ValueError('Invalid configruation value for environment path. Please check the entry for "env" under'
                         'section [PATH] in the config file.')
    caiman_path = sys_cfg['PATHS']['caiman']
    n_processes = sys_cfg['HARDWARE']['n_processes']

    if not os.path.isdir(workdir):
        try:
            os.makedirs(workdir)
        except PermissionError:
            raise PermissionError('You do not appear to have permission to write to the chosen working directory.')

    elif os.path.isfile(workdir):
            raise FileExistsError("Choose a different working dir path")

    if args_str is None:
        args_str = ''
    else:
        args_str = ' ' + args_str

    with open(sh_file, 'w') as sf:
        sf.write('#!/bin/bash\n' +
                 env_activation + '\n' +
                 'export PYTHONPATH="' + caiman_path + '"\n' +
                 'export MKL_NUM_THREADS=1\n' +
                 'export OPENBLAS_NUM_THREADS=1\n'
                 'export MESMERIZE_N_PROCESSES=' + str(n_processes) + '\n'
                 'export USE_CUDA=' + sys_cfg['HARDWARE']['USE_CUDA'] + '\n' +
                 pre_run +
                 'python ' + module_path + args_str + '\n' +
                 post_run
                 )

    st = os.stat(sh_file)
    os.chmod(sh_file, st.st_mode | S_IEXEC)

    return sh_file



# -*- coding: utf-8 -*-
"""
@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007

Adapted from the demo notebook from @agiovann and @epnev
"""

from ipyparallel import Client
import logging
import numpy as np
import os
from scipy.ndimage.filters import gaussian_filter
import sys

import caiman as cm
import caiman.source_extraction.cnmf as cnmf
from caiman.components_evaluation import evaluate_components, estimate_components_quality_auto
from caiman.cluster import setup_cluster
from caiman.paths import caiman_datadir
from caiman.utils.utils import load_dict_from_hdf5

from time import time
import pickle
import traceback
from glob import glob
import json
import shutil

# must be within this block, else windows gives issues
if __name__ == '__main__':
    from mesmerize.common.utils import HdfTools


if not sys.argv[0] == __file__:
    from ...core import ViewerUtils, ViewerWorkEnv


def run(work_dir: str, UUID: str, save_temp_files: str):
    logging.basicConfig(stream=sys.stdout, level=logging.DEBUG,
                        format="%(relativeCreated)12d [%(filename)s:%(funcName)20s():%(lineno)s] [%(process)d] %(message)s")

    start_time = time()

    batch_dir = os.environ['CURR_BATCH_DIR']
    save_temp_files = bool(int(save_temp_files))

    output = {'status': 0, 'output_info': ''}
    n_processes = int(os.environ['_MESMERIZE_N_THREADS'])

    filepath = os.path.join(work_dir, UUID)

    imgpath = f'{filepath}_input.tiff'
    input_params = pickle.load(open(f'{filepath}.params', 'rb'))

    print('******** Creating process pool *********')
    c, dview, n_processes = setup_cluster(
        backend='local', n_processes=n_processes, single_thread=False, ignore_preexisting=True
    )

    try:
        if input_params['use_memmap']:
            memmap_uuid = input_params['memmap_uuid']

            memmap_batchdir = glob(os.path.join(batch_dir, f'memmap-{memmap_uuid}*.mmap'))

            # Check batch dir
            if len(memmap_batchdir) > 0:
                memmap_path = memmap_batchdir[0]
                print(f'********** Found existing memmap in batch dir: {memmap_path} ********** ')

                # copy to work dir
                if not os.path.samefile(batch_dir, work_dir):
                    print('**** Copying memmap to work dir ****')
                    shutil.copy(memmap_path, work_dir)
                    memmap_path = glob(os.path.join(work_dir, f'memmap-{memmap_uuid}*.mmap'))[0]

            else:
                # remake the memmap with the same UUID so that future batch items can rely on it
                print('********** Memmap not found, re-making memmap with the same UUID **********')
                memmap_path = cm.save_memmap(
                    [imgpath], base_name=f'memmap-{memmap_uuid}', is_3D=True, order='C', dview=dview
                )

        else:
            print('********** Making memmap **********')
            memmap_path = cm.save_memmap(
                [imgpath], base_name=f'memmap-{UUID}', is_3D=True, order='C', dview=dview
            )

        print(f'Using memmap:\n{memmap_path}')

        print('********** Loading memmap **********')
        Yr, dims, T = cm.load_memmap(memmap_path)
        Y = np.reshape(Yr, dims + (T,), order='F')

        images = np.reshape(Yr.T, [T] + list(dims), order='F')

        Ain = None

        # seed components
        if 'use_seeds' in input_params.keys():
            if input_params['use_seeds']:
                try:
                    # see if it's an h5 file produced by the nuset_segment GUI
                    hdict = HdfTools.load_dict(
                        os.path.join(f'{filepath}.ain'),
                        'data'
                    )
                    Ain = hdict['sparse_mask']
                except:
                    try:
                        Ain = np.load(f'{UUID}.ain')
                    except Exception as e:
                        output['warnings'] = f'Could not seed components, make sure that ' \
                            f'the .ain file exists in the batch dir: {e}'

        if input_params['use_patches']:
            cnm = cnmf.CNMF(
                n_processes=n_processes,
                dview=dview,
                only_init_patch=True,
                Ain=Ain,
                **input_params['cnmf_kwargs']
            )

        else:
            cnm = cnmf.CNMF(
                n_processes,
                dview=dview,
                Ain=Ain,
                **input_params['cnmf_kwargs']
            )

        cnm.fit(images)

        print('Number of components:' + str(cnm.estimates.A.shape[-1]))

        cnm.params.change_params(
            params_dict={
                **input_params['eval_kwargs'],
                'use_cnn': False
            }
        )

        cnm.estimates.evaluate_components(images, cnm.params, dview=dview)

        if input_params['refit']:
            cnm.params.set('temporal', {'p': input_params['cnmf_kwargs']['p']})
            cnm_ = cnm.refit(images)
        else:
            cnm_ = cnm

        out_filename = f'{UUID}_results.hdf5'
        cnm_.save(out_filename)

        output_files = [out_filename]

        # Save the memmap
        if save_temp_files:
            print("***** Keeping memmap file *****")

            # copy to batch dir if batch_dir != work_dir
            if not os.path.samefile(batch_dir, work_dir):
                print("***** Copying memmap file to batch dir *****")
                shutil.copy(memmap_path, batch_dir)

        # Delete the memmap from the work dir
        if not os.path.samefile(batch_dir, work_dir):
            print("***** Deleting memmap files from work dir *****")
            try:
                os.remove(memmap_path)
            except:
                pass

        output.update(
            {
                'output': UUID,
                'status': 1,
                'output_files': output_files,
                'saved_memmap': save_temp_files
            }
        )

    except Exception as e:
        output.update({'status': 0, 'output_info': traceback.format_exc()})

    cm.stop_server(dview=dview)

    end_time = time()
    processing_time = (end_time - start_time) / 60
    output.update({'processing_time': processing_time})

    json.dump(output, open(filepath + '.out', 'w'))


class Output:
    def __init__(self, batch_path, UUID, viewer_ref):
        vi = ViewerUtils(viewer_ref)

        if not vi.discard_workEnv():
            return

        vi.viewer.status_bar_label.showMessage('Please wait, loading image data...')

        pik_path = os.path.join(batch_path, f'{UUID}_input.pik')
        tiff_path = os.path.join(batch_path, f'{UUID}_input.tiff')
        vi.viewer.workEnv = ViewerWorkEnv.from_pickle(pik_path, tiff_path)

        vi.update_workEnv()
        vi.viewer.status_bar_label.showMessage('Finished loading image data...')

        input_params_path = os.path.join(batch_path, f'{UUID}.params')
        input_params = pickle.load(open(input_params_path, 'rb'))

        cnmf_data = load_dict_from_hdf5(
            os.path.join(batch_path, f'{UUID}_results.hdf5')
        )

        roi_manager_gui = vi.viewer.parent().get_module('roi_manager')
        roi_manager_gui.start_backend('VolCNMF')

        roi_manager_gui.manager.add_all_components(
            cnmf_data,
            input_params_dict=input_params,
        )

        name = input_params['item_name']
        vi.viewer.ui.label_curr_img_seq_name.setText(f'CNMF 3D: {name}')
        vi.viewer.workEnv.history_trace.append({'cnmf_3d': input_params})
        vi.enable_ui(True)


#if sys.argv[0] == __file__:
if __name__ == '__main__':
    run(*sys.argv[1:])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on June 18 2018

@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

from PyQt5 import QtCore, QtGui, QtWidgets
from .main_widget import ProjectBrowserWidget
from ...pyqtgraphCore.console import ConsoleWidget
from .pytemplates.mainwindow_pytemplate import Ui_MainWindow
from spyder.widgets.variableexplorer import objecteditor
import pandas as pd
import os
from ...common import configuration, get_project_manager
from functools import partial


class ProjectBrowserWindow(QtWidgets.QMainWindow):
    def __init__(self, parent=None):
        QtWidgets.QMainWindow.__init__(self, parent)

        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)

        self.project_browser = ProjectBrowserWidget(self, get_project_manager().dataframe)
        self.current_tab = 0
        self.project_browser.tab_widget.currentChanged.connect(self.set_current_tab)
        self.setCentralWidget(self.project_browser)

        self.ui.actionto_pickle.triggered.connect(partial(self.save_path_dialog, file_ext='pikl', save_root=True))
        self.ui.actionto_csv.triggered.connect(partial(self.save_path_dialog, file_ext='csv', save_root=True))
        self.ui.actionto_excel.triggered.connect(partial(self.save_path_dialog, file_ext='xlsx', save_root=True))

        self.ui.actionto_pickle_tab.triggered.connect(partial(self.save_path_dialog, file_ext='pikl', save_root=False))
        self.ui.actionto_csv_tab.triggered.connect(partial(self.save_path_dialog, file_ext='csv', save_root=False))
        self.ui.actionto_excel_tab.triggered.connect(partial(self.save_path_dialog, file_ext='xlsx', save_root=False))

        self.ui.actionDataframe_editor.triggered.connect(self.edit_root_dataframe)

        self.ui.actionViewCurrent_dataframe.triggered.connect(self.view_current_dataframe)
        self.ui.actionCurrent_tab_filter_history.triggered.connect(self.view_filters)

        self.ui.actionUpdate_current_tab.triggered.connect(self.update_tab_from_child_dataframe)
        self.ui.actionUpdate_all_tabs.triggered.connect(self.update_tabs_from_child_dataframes)

        self.ui.actionSave_to_project.triggered.connect(self._save_dataframe)

        ns = {'pd': pd,
              'project_browser': self.project_browser,
              'this': self,
              'get_dataframe': lambda: self.get_current_dataframe()[0],
              'get_root_dataframe': lambda: self.get_current_dataframe(get_root=True)[0],
              'set_root_dataframe': lambda df: get_project_manager().set_dataframe(df)
              }

        txt = "Namespaces:          \n" \
              "pandas as pd         \n" \
              "Useful callables:\n" \
              "get_dataframe() - returns dataframe of the current tab\n" \
              "get_root_dataframe() - always return dataframe of root tab (i.e. entire dataframe)\n" \
              "set_root_dataframe() - save the root dataframe\n" \
              "self as this         \n" \

        cmd_history_file = os.path.join(configuration.console_history_path, 'project_browser.pik')

        self.ui.dockConsole.setWidget(ConsoleWidget(namespace=ns, text=txt, historyFile=cmd_history_file))
        # self.resizeDocks([self.ui.dockConsole], [235], QtCore.Qt.Vertical)
        self.ui.dockConsole.hide()

        self._status_bar = None
        self.status_bar = self.statusBar()

        self.setWindowTitle('Project Browser')

    @property
    def status_bar(self) -> QtWidgets.QStatusBar:
        return self._status_bar

    @status_bar.setter
    def status_bar(self, status_bar: QtWidgets.QStatusBar):
        self._status_bar = status_bar

    def set_current_tab(self, ix: int):
        self.current_tab = ix

    def save_path_dialog(self, file_ext: str, save_root=False):
        path = QtWidgets.QFileDialog.getSaveFileName(None, 'Save DataFrame as', '', f'(*.{file_ext})')
        if path == '':
            return
        if path[0].endswith('.' + file_ext):
            path = path[0]
        else:
            path = path[0] + '.' + file_ext

        if file_ext == 'pikl':
            self.get_current_dataframe(get_root=save_root)[0].to_pickle(path, protocol=4)
        elif file_ext == 'csv':
            self.get_current_dataframe(get_root=save_root)[0].to_csv(path)
        elif file_ext == '.xlsx':
            self.get_current_dataframe(get_root=save_root)[0].to_excel(path)

    def get_current_dataframe(self, get_root=False) -> (pd.DataFrame, list):
        if self.current_tab == 0 or get_root:
            dataframe = get_project_manager().get_dataframe()
            return dataframe, []
        else:
            tab_name = self.project_browser.tab_widget.widget(self.current_tab).tab_name
            child = get_project_manager().child_dataframes[tab_name]
        return child['dataframe'], child['filter_history']

    def _save_dataframe(self):
        get_project_manager().save_dataframe()

    def edit_root_dataframe(self):
        QtWidgets.QMessageBox.warning(self, 'Proceed with caution!',
                                      'You are about to edit the project DataFrame, be careful!\n'
                                      'If you make a mistake press "Close"')

        df = objecteditor.oedit(get_project_manager().get_dataframe())
        if df is not None:
            get_project_manager().set_dataframe(df)

    def view_current_dataframe(self):
        objecteditor.oedit(self.get_current_dataframe()[0])

    def view_filters(self):
        filters = self.get_current_dataframe()[1]

        QtWidgets.QMessageBox.information(self, 'Filters for current tab', '\n'.join(filters))

    def reload_all_tabs(self):
        self.status_bar.showMessage('Please wait, loading tabs...')
        for i in range(1, self.project_browser.tab_widget.count()):
            self.project_browser.tab_widget.removeTab(i)

        # for child in configuration.project_manager.child_dataframes.keys():
        #     configuration.project_manager.child_dataframes.pop(child)

        get_project_manager().create_sub_dataframe()

        for child in get_project_manager().child_dataframes.keys():
            dataframe = get_project_manager().child_dataframes[child]['dataframe']
            filter_history = get_project_manager().child_dataframes[child]['filter_history']
            self.project_browser.add_tab(dataframe, filter_history, name=child)
        self.status_bar.showMessage('Finished loading tabs!')

    def update_tab_from_child_dataframe(self, tab_name=None):
        if tab_name is None:
            tab_name = self.project_browser.tab_widget.widget(self.current_tab).tab_name

        child = get_project_manager().child_dataframes[tab_name]
        dataframe = child['dataframe']
        filter_history = child['filter_history']
        self.project_browser.tabs[tab_name].dataframe = dataframe
        self.project_browser.tabs[tab_name].filter_history = filter_history
        self.project_browser.tabs[tab_name].populate_tab()

    def update_tabs_from_child_dataframes(self):
        for tab_name in self.project_browser.tabs.keys():
            if tab_name == 'root':
                continue
            self.update_tab_from_child_dataframe(tab_name)



import numpy as np
from ...common.qdialogs import *
from ..core.common import ViewerUtils
from PyQt5 import QtWidgets
from ..core import ViewerWorkEnv
from ..core.data_types import ImgData
from multiprocessing.shared_memory import SharedMemory
import os
from subprocess import Popen


class Widget(QtWidgets.QWidget):
    def __init__(self, parent):
        QtWidgets.QWidget.__init__(self, parent=parent)
        self.vboxlayout = QtWidgets.QVBoxLayout(self)
        self.hboxlayout = QtWidgets.QHBoxLayout(self)

        self.label_choose_file = QtWidgets.QLabel()
        self.label_choose_file.setText("Choose `.isxd` file")
        self.vboxlayout.addWidget(self.label_choose_file)

        self.line_edit_path = QtWidgets.QLineEdit()
        self.line_edit_path.setPlaceholderText("Path to `.isxd` file")
        self.line_edit_path.returnPressed.connect(self.parent().load_file)
        self.hboxlayout.addWidget(self.line_edit_path)

        self.btn_open_file_dialog = QtWidgets.QPushButton()
        self.btn_open_file_dialog.setText("...")
        self.btn_open_file_dialog.clicked.connect(self.parent().file_dialog)
        self.hboxlayout.addWidget(self.btn_open_file_dialog)

        self.vboxlayout.addLayout(self.hboxlayout)

        self.btn_load_file = QtWidgets.QPushButton()
        self.btn_load_file.setText("Load File")
        self.btn_load_file.clicked.connect(self.parent().load_file)
        self.vboxlayout.addWidget(self.btn_load_file)


class ModuleGUI(QtWidgets.QDockWidget):
    def __init__(self, parent, viewer_reference):
        self.vi = ViewerUtils(viewer_reference)
        QtWidgets.QDockWidget.__init__(self, parent)
        self.setFloating(True)
        self.setWindowTitle('Inscopix Importer')

        self.widget = Widget(parent=self)

        self.setWidget(self.widget)

        # self.setLayout(self.vboxlayout)
    @use_open_file_dialog("Choose .isxd file", "", exts=["*.isxd"])
    def file_dialog(self, path, *args, **kwargs):
        self.widget.line_edit_path.setText(path)

    @present_exceptions("File open error", "The following error occurred when opening the file")
    def load_file(self, *args, **kwargs):
        shm = SharedMemory(create=True, size=5 * np.dtype('<U64').itemsize)

        # get array metadata, shared memory buffer name, dtype, and shape
        array_metadata = np.ndarray(shape=(5,), dtype=np.dtype('<U64'), buffer=shm.buf)

        importer_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), '_inscopix_importer.py')

        file_path = self.widget.line_edit_path.text()

        cmd = ['python', importer_path, '--isx-path', file_path, '--shm-meta-array-name', shm.name]
        print(cmd)
        proc = Popen(
            cmd,
            env=os.environ.copy()
        )
        proc.wait()

        # read the metadata
        name = array_metadata[0]
        dtype = array_metadata[1]
        shape = array_metadata[2]

        # open the shared buffer
        existing_shm = SharedMemory(name=array_metadata[0])

        shape = tuple(map(int, shape.split(',')))

        _imgseq = np.ndarray(shape, dtype=np.dtype(dtype), buffer=existing_shm.buf)

        imgseq = np.zeros(shape=shape, dtype=np.dtype(dtype))
        imgseq[:] = _imgseq[:]

        d = \
            {
                'fps': float(array_metadata[3]),
                'origin': array_metadata[4],
                'date': '00000000_000000'
            }

        imgdata = ImgData(imgseq, d)

        self.vi.viewer.workEnv = ViewerWorkEnv(imgdata)
        self.vi.update_workEnv()

        existing_shm.close()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on June 14 2018

@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

from collections import UserList
from PyQt5 import QtCore, QtGui, QtWidgets
from . import start
from functools import partial
import os


mpath = os.path.abspath(__file__)
mdir = os.path.dirname(mpath)


class WindowManager:
    """"""
    def __init__(self):
        self.welcome_window = None
        self.project_browser = None
        self.viewers = WindowList('Viewer')
        self.flowcharts = WindowList('Flowchart')

    def get_batch_manager(self, run_batch: list = None) -> QtWidgets.QWidget:
        return self.welcome_window.get_batch_manager(run_batch)

    def get_project_browser(self):
        if self.project_browser is None:
            raise AttributeError('Project Browser is not running')
        return self.project_browser

    # def initalize(self):
        # self.project_browsers = WindowClass('Project Browser')

    def get_new_viewer_window(self):
        w = start.viewer()
        self.viewers.append(w)
        w.destroyed.connect(partial(self.viewers.remove_window, w))
        return w

    def get_new_flowchart(self, filename: str = None, parent = None):
        w = start.flowchart(filename, parent)
        self.flowcharts.append(w)
        w.destroyed.connect(partial(self.flowcharts.remove_window, w))
        return w


class WindowList(UserList):
    """Simple management of the main windows of a certain type
     and the associated windows list widget on the welcome window"""
    def __init__(self, window_name: str):
        super(WindowList, self).__init__()
        self.window_name = window_name
        # self._selected_window = None
        #
        # self.list_widget = QtWidgets.QListWidget()
        # self.list_widget.itemDoubleClicked.connect(self._show_window)
        # self.list_widget.currentItemChanged.connect(self._set_selected_window)
        # self.list_widget.setToolTip('Double click to show window')
        #
        # action_delete_column = QtWidgets.QWidgetAction(self)
        # action_delete_column.setText('Delete column')
        # action_delete_column.triggered.connect(self.delete_column)

    def append(self, window: QtWidgets.QMainWindow):
        # self.list_widget.addItem(str(self.__len__() + 1))
        window.setWindowTitle('{} - {}'.format(self.window_name, self.__len__() + 1))
        window.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        window.setWindowIcon(QtGui.QIcon(mdir + '/icons/main_icon.gif'))
        super(WindowList, self).append(window)

    # def _get_selected_window(self) -> QtWidgets.QMainWindow:
    #     return self._selected_window

    # def _set_selected_window(self):
        # i = self.list_widget.currentRow()
        # self._selected_window = self.__getitem__(i)

    # def _show_window(self):
    #     self._selected_window.show()

    def remove_window(self, window: QtWidgets.QMainWindow):
        self.data.remove(window)
        for ix, window in enumerate(self.data):
            window.setWindowTitle('{} - {}'.format(self.window_name, ix + 1))

    #
    # def __getitem__(self, item) -> QtWidgets.QMainWindow:
    #     return super(WindowClass, self).__getitem__(item)





#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#@author: kushal

#Chatzigeorgiou Group
#Sars International Centre for Marine Molecular Biology

#GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007

from ..base import BasePlotWidget
from ...utils import *
from ....pyqtgraphCore.widgets.MatplotlibWidget import MatplotlibWidget
from ....pyqtgraphCore.console import ConsoleWidget
from ....common.configuration import console_history_path
from ....common.qdialogs import *
from matplotlib.axes import Axes
from matplotlib.patches import Polygon, Patch
from typing import *
import tifffile
import os
import pandas as pd
from ....analysis import Transmission
from .control_widget import Ui_Controls


class ControlDock(QtWidgets.QDockWidget):
    sig_changed = QtCore.pyqtSignal()

    def __init__(self, parent):
        QtWidgets.QDockWidget.__init__(self, parent=parent)
        self.ui = Ui_Controls()
        self.ui.setupUi(self)

        self.widget_registry = WidgetRegistry()

        self.ui.cmap_img.set_cmap('inferno')
        self.ui.cmap_patches.set_cmap('tab10')

        self.widget_registry.register(self.ui.combo_categorical_column,
                                      setter=self.ui.combo_categorical_column.setText,
                                      getter=self.ui.combo_categorical_column.currentText,
                                      name='categorical_column')

        self.widget_registry.register(self.ui.cmap_img,
                                      setter=self.ui.cmap_img.set_cmap,
                                      getter=self.ui.cmap_img.get_cmap,
                                      name='cmap_img')

        self.widget_registry.register(self.ui.cmap_patches,
                                      setter=self.ui.cmap_patches.set_cmap,
                                      getter=self.ui.cmap_patches.get_cmap,
                                      name='cmap_patches')

        self.widget_registry.register(self.ui.radioButtonMax,
                                      setter=self.ui.radioButtonMax.setChecked,
                                      getter=self.ui.radioButtonMax.isChecked,
                                      name='max_projection')

        self.widget_registry.register(self.ui.radioButtonStd,
                                      setter=self.ui.radioButtonStd.setChecked,
                                      getter=self.ui.radioButtonStd.isChecked,
                                      name='std_projection')

        self.widget_registry.register(self.ui.list_widget_samples,
                                      setter=lambda l: self.ui.list_widget_samples.setSelectedItems(l),
                                      getter=lambda: [self.ui.list_widget_samples.currentItem().text()],
                                      name='selected_sample')

        self.widget_registry.register(self.ui.checkBoxFill,
                                      setter=self.ui.checkBoxFill.setChecked,
                                      getter=self.ui.checkBoxFill.isChecked,
                                      name='fill_patches')

        self.widget_registry.register(self.ui.doubleSpinBoxLineWidth,
                                      setter=self.ui.doubleSpinBoxLineWidth.setValue,
                                      getter=self.ui.doubleSpinBoxLineWidth.value,
                                      name='line_width')

        self.widget_registry.register(self.ui.doubleSpinBoxAlpha,
                                      setter=self.ui.doubleSpinBoxAlpha.setValue,
                                      getter=self.ui.doubleSpinBoxAlpha.value,
                                      name='alpha')

        self.ui.combo_categorical_column.currentTextChanged.connect(self.sig_changed)
        self.ui.cmap_img.currentItemChanged.connect(self.sig_changed)
        self.ui.cmap_patches.currentItemChanged.connect(self.sig_changed)
        self.ui.radioButtonMax.clicked.connect(self.sig_changed)
        self.ui.radioButtonStd.clicked.connect(self.sig_changed)
        self.ui.checkBoxFill.clicked.connect(self.sig_changed)
        self.ui.doubleSpinBoxLineWidth.valueChanged.connect(self.sig_changed)
        self.ui.doubleSpinBoxAlpha.valueChanged.connect(self.sig_changed)
        self.ui.list_widget_samples.currentItemChanged.connect(self.sig_changed)

    def fill_widget(self, samples: list, categorical_columns: list):
        self.ui.combo_categorical_column.setItems(categorical_columns)
        self.ui.list_widget_samples.setItems(samples)

    def get_state(self) -> dict:
        return self.widget_registry.get_state()

    def set_state(self, state: dict):
        self.widget_registry.set_state(state)


class PlotArea(MatplotlibWidget):
    def __init__(self):
        MatplotlibWidget.__init__(self)
        self.error_label = QtWidgets.QLabel(self)
        self.error_label.setMaximumHeight(32)
        self.vbox.addWidget(self.error_label)
        self._ax = None

    @property
    def ax(self) -> Axes:
        if self._ax is None:
            raise ValueError('Axes not set')

        return self._ax

    def clear(self):
        self.fig.clear()
        self._ax = self.fig.add_subplot(111)


class SpaceMapWidget(QtWidgets.QMainWindow, BasePlotWidget):
    drop_opts = []

    def __init__(self):
        QtWidgets.QMainWindow.__init__(self, parent=None)
        BasePlotWidget.__init__(self)

        self.setWindowTitle('Space Map Plot')

        self.plot = PlotArea()
        self.setCentralWidget(self.plot)

        self.control_widget = ControlDock(self)
        self.addDockWidget(QtCore.Qt.LeftDockWidgetArea, self.control_widget)
        self.control_widget.sig_changed.connect(lambda: self.update_plot())

        self.update_live = self.control_widget.ui.checkBoxLiveUpdate.isChecked()
        self.control_widget.ui.checkBoxLiveUpdate.toggled.connect(self.set_update_live)

        self.control_widget.ui.pushButtonUpdatePlot.clicked.connect(self.update_plot)

        self.control_widget.ui.pushButtonSave.clicked.connect(self.save_plot_dialog)
        self.control_widget.ui.pushButtonLoad.clicked.connect(self.open_plot_dialog)

        self.control_widget.ui.groupBox.setVisible(False)

        cmd_history_file = os.path.join(console_history_path, 'space_map.pik')

        ns = {'this': self,
              'get_plot': lambda: self.plot,
              }

        txt = ["Namespaces",
               "self as 'this'",
               "Useful callable: get_plot()"
               ]

        txt = "\n".join(txt)

        self.console = ConsoleWidget(parent=self, namespace=ns, text=txt, historyFile=cmd_history_file)

        self.dockConsole = QtWidgets.QDockWidget(self)
        self.dockConsole.setWindowTitle('Console')
        self.dockConsole.setFeatures(
            QtWidgets.QDockWidget.DockWidgetFloatable | QtWidgets.QDockWidget.DockWidgetMovable)
        self.dockConsole.setWidget(self.console)
        self.addDockWidget(QtCore.Qt.BottomDockWidgetArea, self.dockConsole)

        self.block_signals_list = [self.control_widget]

        self.error_label = self.plot.error_label
        self.exception_holder = None
        self.error_label.mousePressEvent = self.show_exception_info

        self.sample_df = None  #: sub-dataframe of the current sample

        self.img: np.ndarray = np.empty(0)
        self.control_widget.ui.verticalSliderZLevel.valueChanged.connect(self._update_img)
        self.previous_sample_id: str = None

    def set_update_live(self, b: bool):
        self.control_widget.ui.checkBoxLiveUpdate.setChecked(b)
        self.update_live = b

    @BasePlotWidget.signal_blocker
    def set_input(self, transmission: Transmission):
        """Set the input transmission"""
        if (self._transmission is None) or self.update_live:
            super(SpaceMapWidget, self).set_input(transmission)
            self.update_plot()

    @BasePlotWidget.signal_blocker
    def fill_control_widget(self, data_columns: list, categorical_columns: list, uuid_columns: list):
        samples = list(self.transmission.df['SampleID'].unique())
        self.control_widget.fill_widget(samples=samples, categorical_columns=categorical_columns)

    @exceptions_label('error_label', 'exception_holder', 'Error while setting data', 'Make sure you have selected appropriate columns')
    def update_plot(self, *args, **kwargs):
        self.plot.clear()
        self.error_label.clear()

        plot_opts = self.control_widget.widget_registry.get_state()

        if plot_opts['max_projection']:
            projection = 'max'
        elif plot_opts['std_projection']:
            projection = 'std'

        sample_id = plot_opts['selected_sample'][0]

        self.plot.ax.set_title(sample_id)

        self.sample_df = self.transmission.df[self.transmission.df['SampleID'] == sample_id]

        self.img = self.load_image(projection)

        if self.img.ndim > 2:
            self.control_widget.ui.groupBox.setVisible(True)
            self.control_widget.ui.verticalSliderZLevel.setMaximum(self.img.shape[0])
            self.control_widget.ui.spinBoxZLevel.setMaximum(self.img.shape[0])
            # self.control_widget.ui.spinBoxZLevel.setValue(0)
            if self.previous_sample_id == sample_id:
                self._update_img(
                    self.control_widget.ui.verticalSliderZLevel.value()
                )
            else:
                self.control_widget.ui.verticalSliderZLevel.setValue(0)
                self._update_img(0)
                self.previous_sample_id = sample_id
        else:
            self.control_widget.ui.groupBox.setVisible(False)
            self._update_img()

    def _update_img(self, zlevel: int = 0):
        self.plot.ax.cla()
        plot_opts = self.control_widget.widget_registry.get_state()

        categorical_column = plot_opts['categorical_column']

        if len(plot_opts['selected_sample']) < 1:
            raise ValueError('No Sample selected')

        cmap_img = plot_opts['cmap_img']
        cmap_patches = plot_opts['cmap_patches']
        fill_patches = plot_opts['fill_patches']
        line_width = plot_opts['line_width']
        alpha = plot_opts['alpha']

        labels = self.sample_df[categorical_column]
        cmap_labels = get_colormap(labels=labels.unique(), cmap=cmap_patches)

        if self.img.ndim > 2:
            vmax = np.nanmedian(self.img[zlevel, :, :]) + (10 * np.nanstd(self.img))
            self.plot.ax.imshow(
                self.img[zlevel, :, :].transpose(1, 0),
                origin='lower',cmap
                =cmap_img,
                vmax=vmax
            )
        else:
            vmax = np.nanmedian(self.img) + (10 * np.nanstd(self.img))
            self.plot.ax.imshow(self.img.transpose(1, 0), origin='lower', cmap=cmap_img, vmax=vmax)

        for ix, r in self.sample_df.iterrows():
            roi = r['ROI_State']

            if roi['roi_type'] == 'VolMultiCNMFROI':
                if not roi['zcenter'] == zlevel:
                    continue

            if roi['roi_type'] == 'VolCNMF':
                coors3d = roi['coors'][zlevel]
                current_coors = coors3d[~np.isnan(coors3d).any(axis=1)]
                if not current_coors.size > 0:  # roi not visible in current plane
                    continue

                xs = current_coors[:, 1].flatten().astype(int)
                ys = current_coors[:, 0].flatten().astype(int)
            else:
                xs = roi['roi_xs']
                ys = roi['roi_ys']

            cors = np.dstack([xs, ys])[0]
            label = r[categorical_column]

            poly = Polygon(cors, fill=fill_patches, color=cmap_labels[label], linewidth=line_width, alpha=alpha)

            self.plot.ax.add_patch(poly)

        handles = [Patch(color=cmap_labels[k], label=k) for k in cmap_labels.keys()]
        self.plot.ax.legend(handles=handles, ncol=int(np.sqrt(len(handles))), loc='lower right', title=categorical_column)

        self.plot.fig.tight_layout()

        self.plot.draw()

    def load_image(self, projection: str):
        img_uuid = self.sample_df['ImgUUID'].iloc[0]
        sample_id = self.sample_df['SampleID'].iloc[0]

        if isinstance(img_uuid, pd.Series):
            img_uuid = img_uuid.item()

        if not isinstance(img_uuid, str):
            raise ValueError('Datatype for Projection Path must be pandas.Series or str, it is currently : ' + str(
                type(img_uuid)))

        # see if data are 3D
        if os.path.isfile(
            os.path.join(self.transmission.get_proj_path(), 'images', f'{sample_id}-_-{img_uuid}_max_proj-0.tiff')
        ):
            z = 0
            imgs = []
            while True:
                img_path = os.path.join(self.transmission.get_proj_path(), 'images', f'{sample_id}-_-{img_uuid}_{projection}_proj-{z}.tiff')
                if os.path.isfile(img_path):
                    imgs.append(tifffile.TiffFile(img_path).asarray())
                    z += 1
                else:
                    break

            img = np.stack(imgs)
        else:
            img_path = os.path.join(self.transmission.get_proj_path(), 'images', f'{sample_id}-_-{img_uuid}_{projection}_proj.tiff')
            img = tifffile.TiffFile(img_path).asarray()

        return img

    @BasePlotWidget.signal_blocker
    def set_update_live(self, b: bool):
        pass

    def get_plot_opts(self, drop: bool = False) -> dict:
        return self.control_widget.get_state()

    @BasePlotWidget.signal_blocker
    def set_plot_opts(self, opts: dict):
        self.control_widget.set_state(opts)

    def show_exception_info(self, mouse_press_ev):
        if self.exception_holder is not None:
            QtWidgets.QMessageBox.warning(self, *self.exception_holder)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#@author: kushal

#Chatzigeorgiou Group
#Sars International Centre for Marine Molecular Biology

#GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007


import numpy as np
import pandas as pd
from typing import *
from . import Transmission
from itertools import product


def get_proportions(xs: Union[pd.Series, np.ndarray, list], ys: Union[pd.Series, np.ndarray, pd.Series],
                    xs_name: str = 'xs', ys_name: str = 'ys',
                    swap: bool = False, percentages: bool = True) -> pd.DataFrame:
    """
    Get the proportions of xs vs ys.

    xs & ys are categorical data.

    :param xs: data plotted on the x axis
    :type xs: Union[pd.Series, np.ndarray]

    :param ys: proportions of unique elements in ys are calculated per xs
    :type ys: Union[pd.Series, np.ndarray]

    :param xs_name: name for the xs data, useful for labeling the axis in plots
    :type xs_name: str

    :param ys_name: name for the ys data, useful for labeling the axis in plots
    :type ys_name: str

    :param swap: swap x and y
    :type swap: bool

    :return:   DataFrame that can be plotted in a proportions bar graph
    :rtype: pd.DataFrame
    """

    if len(xs) != len(ys):
        raise ValueError('Length of xs and ys must match exactly')

    if isinstance(xs, np.ndarray):
        if xs.ndim > 1:
            raise ValueError('Can only accept 1D numpy array')

    if isinstance(ys, np.ndarray):
        if ys.ndim > 1:
            raise ValueError('Can only accept 1D numpy array')

    if swap:
        xs, ys = ys, xs
        xs_name, ys_name = ys_name, xs_name

    df = pd.DataFrame({xs_name: xs, ys_name: ys})
    if percentages:
        props_df = df.groupby([xs_name, ys_name]).agg({ys_name: 'count'}).groupby(by=xs_name).apply(lambda x: (x / x.sum()) * 100).unstack()
        props_df.columns = props_df.columns.get_level_values(-1)
    else:
        props_df = df.groupby([xs_name, ys_name]).agg({ys_name: 'count'}).unstack()
        props_df.columns = props_df.columns.get_level_values(-1)

    return props_df


def get_sampling_rate(transmission: Transmission, tolerance: Optional[float] = 0.1) -> float:
    """
    Returns the mean sampling rate of all data in a Transmission if it is within the specified tolerance. Otherwise throws an exception.

    :param transmission:    Transmission object of the data from which sampling rate is obtained.
    :type transmission:     Transmission

    :param tolerance:       Maximum tolerance (in Hertz) of sampling rate variation between different samples
    :type tolerance:        float

    :return:                The mean sampling rate of all data in the Transmission
    :rtype:                 float
    """
    sampling_rates = []
    for db in transmission.history_trace.data_blocks:
        if transmission.history_trace.check_operation_exists(db, 'resample'):
            sampling_rates.append(transmission.history_trace.get_operation_params(db, 'resample')['output_rate'])
        else:
            r = pd.DataFrame(transmission.get_data_block_dataframe(db).meta.to_list())['fps'].unique()
            # if rates.size > 1:
            #     raise ValueError("Sampling rates for the data do not match")
            # else:
            sampling_rates.append(r)

    rates = np.hstack([sampling_rates])

    if np.ptp(rates) > tolerance:
        raise ValueError("Sampling rates of the data differ by "
                         "greater than the set tolerance of " + str(tolerance) + " Hz")

    framerate = float(np.mean(sampling_rates))

    return framerate


def get_frequency_linspace(transmission: Transmission) -> Tuple[np.ndarray, float]:
    """
    Get the frequency linspace.

    Throwns an exception if all datablocks do not have the same linspace & Nyquist frequencies

    :param transmission: Transmission containing data from which to get frequency linspace

    :return: tuple: (frequency linspace as a 1D numpy array, nyquist frequency)
    :rtype:  Tuple[np.ndarray, float]
    """

    # Check that all sampling rates are equal
    get_sampling_rate(transmission)

    fs = []
    nqs = []

    for db in transmission.history_trace.data_blocks:
        params = transmission.history_trace.get_operation_params(db, 'rfft')

        f = params['frequencies']
        fs.append(np.array(f))

        nqs.append(params['nyquist_frequency'])

    if len(set(nqs)) > 1:
        raise ValueError("Nyquist frequency of all data blocks must match exactly")

    # Check that the discrete frequencies of all datablocks match exactly
    for i, j in product(*(range(len(fs)), )*2):
        if i == j:
            continue
        if not np.array_equal(fs[i], fs[j]):
            raise ValueError("Discrete frequencies of all data blocks must match exactly")

    return fs[0], nqs[0]


def get_array_size(transmission: Transmission, data_column: str) -> int:
    """Returns the size of the 1D arrays in the specified data column. Throws an exception if they do not match

    :param transmission:    Desired Transmission
    :param data_column:     Data column of the Transmission from which to retrieve the size

    :type transmission:     Transmission
    :type data_column:      str

    :return:                Size of the 1D arrays of the specified data column
    :rtype:                 int
    """
    if data_column not in transmission.df.columns:
        raise KeyError("Requested data column: " + data_column + " not found in transmission DataFrame")
    array_size = transmission.df[data_column].apply(lambda a: a.size).unique()

    if array_size.size > 1:
        raise ValueError("Size of all arrays in data column must match exactly.")

    return array_size[0]


def organize_dataframe_columns(columns: Iterable[str]) -> Tuple[List[str], List[str], List[str]]:
    """
    Organizes DataFrame columns into data column, categorical label columns, and uuid columns.

    :param columns: All DataFrame columns

    :return:        (data_columns, categorical_columns, uuid_columns)
    :rtype:         Tuple[List[str], List[str], List[str]]
    """
    columns = list(columns)
    if '_BLOCK_' in columns:
        columns.remove('_BLOCK_')

    dcols = [c for c in columns if c.startswith('_')]

    ucols = [c for c in columns if ('uuid' in c) or ('UUID' in c)]

    ccols = [c for c in columns if (not c.startswith('_')) and
             (c not in ['CurvePath', 'ImgPath', 'ImgUUID', 'ROI_State', 'meta', 'ImgInfoPath', 'misc'])
             ]

    dcols.sort();ccols.sort();ucols.sort()

    return dcols, ccols, ucols


def pad_arrays(a: np.ndarray, method: str = 'random', output_size: int = None, mode: str = 'minimum',
               constant: Any = None) -> np.ndarray:
    """
    Pad all the input arrays so that are of the same length. The length is determined by the largest input array.
    The padding value for each input array is the minimum value in that array.

    Padding for each input array is either done after the array's last index to fill up to the length of the
    largest input array (method 'fill-size') or the padding is randomly flanked to the input array (method 'random')
    for easier visualization.

    :param a: 1D array where each element is a 1D array
    :type a: np.ndarray

    :param method: one of 'fill-size' or 'random', see docstring for details
    :type method: str

    :param output_size: not used

    :param mode: one of either 'constant' or 'minimum'.
                 If 'minimum' the min value of the array is used as the padding value.
                 If 'constant' the values passed to the "constant" argument is used as the padding value.
    :type mode: str

    :param constant: padding value if 'mode' is set to 'constant'
    :type constant: Any

    :return: Arrays padded according to the chosen method. 2D array of shape [n_arrays, size of largest input array]
    :rtype: np.ndarray
    """

    l = 0  # size of largest time series

    # Get size of largest time series
    for c in a:
        s = c.size
        if s > l:
            l = s

    if (output_size is not None) and (output_size < l):
        raise ValueError('Output size must be equal to larger than the size of the largest input array')

    # pre-allocate output array
    p = np.zeros(shape=(a.size, l), dtype=a[0].dtype)

    # pad each 1D time series
    for i in range(p.shape[0]):
        s = a[i].size

        if s == l:
            p[i, :] = a[i]
            continue

        max_pad_en_ix = l - s

        if method == 'random':
            pre = np.random.randint(0, max_pad_en_ix)
        elif method == 'fill-size':
            pre = 0
        else:
            raise ValueError('Must specific method as either "random" or "fill-size"')

        post = l - (pre + s)

        if mode == 'constant':
            p[i, :] = np.pad(a[i], (pre, post), 'constant', constant_values=constant)
        else:
            p[i, :] = np.pad(a[i], (pre, post), 'minimum')

    return p



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'welcome_window_pytemplate.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(715, 560)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(MainWindow.sizePolicy().hasHeightForWidth())
        MainWindow.setSizePolicy(sizePolicy)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout_7 = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout_7.setObjectName("verticalLayout_7")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.label_4 = QtWidgets.QLabel(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_4.sizePolicy().hasHeightForWidth())
        self.label_4.setSizePolicy(sizePolicy)
        self.label_4.setObjectName("label_4")
        self.verticalLayout_3.addWidget(self.label_4)
        self.btnViewer = QtWidgets.QPushButton(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.btnViewer.sizePolicy().hasHeightForWidth())
        self.btnViewer.setSizePolicy(sizePolicy)
        self.btnViewer.setMinimumSize(QtCore.QSize(131, 118))
        self.btnViewer.setMaximumSize(QtCore.QSize(131, 118))
        self.btnViewer.setAutoFillBackground(True)
        self.btnViewer.setText("")
        self.btnViewer.setFlat(False)
        self.btnViewer.setObjectName("btnViewer")
        self.verticalLayout_3.addWidget(self.btnViewer)
        self.horizontalLayout.addLayout(self.verticalLayout_3)
        self.verticalLayout_4 = QtWidgets.QVBoxLayout()
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.label_5 = QtWidgets.QLabel(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_5.sizePolicy().hasHeightForWidth())
        self.label_5.setSizePolicy(sizePolicy)
        self.label_5.setObjectName("label_5")
        self.verticalLayout_4.addWidget(self.label_5)
        self.btnFlowchart = QtWidgets.QPushButton(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.btnFlowchart.sizePolicy().hasHeightForWidth())
        self.btnFlowchart.setSizePolicy(sizePolicy)
        self.btnFlowchart.setMinimumSize(QtCore.QSize(131, 118))
        self.btnFlowchart.setMaximumSize(QtCore.QSize(131, 118))
        self.btnFlowchart.setAutoFillBackground(True)
        self.btnFlowchart.setText("")
        self.btnFlowchart.setFlat(False)
        self.btnFlowchart.setObjectName("btnFlowchart")
        self.verticalLayout_4.addWidget(self.btnFlowchart)
        self.horizontalLayout.addLayout(self.verticalLayout_4)
        self.verticalLayoutProjectBrowser = QtWidgets.QVBoxLayout()
        self.verticalLayoutProjectBrowser.setObjectName("verticalLayoutProjectBrowser")
        self.labelProjectBrowser = QtWidgets.QLabel(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.labelProjectBrowser.sizePolicy().hasHeightForWidth())
        self.labelProjectBrowser.setSizePolicy(sizePolicy)
        self.labelProjectBrowser.setObjectName("labelProjectBrowser")
        self.verticalLayoutProjectBrowser.addWidget(self.labelProjectBrowser)
        self.btnProjectBrowser = QtWidgets.QPushButton(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.btnProjectBrowser.sizePolicy().hasHeightForWidth())
        self.btnProjectBrowser.setSizePolicy(sizePolicy)
        self.btnProjectBrowser.setMinimumSize(QtCore.QSize(131, 118))
        self.btnProjectBrowser.setMaximumSize(QtCore.QSize(131, 118))
        self.btnProjectBrowser.setAutoFillBackground(True)
        self.btnProjectBrowser.setText("")
        self.btnProjectBrowser.setFlat(False)
        self.btnProjectBrowser.setObjectName("btnProjectBrowser")
        self.verticalLayoutProjectBrowser.addWidget(self.btnProjectBrowser)
        self.horizontalLayout.addLayout(self.verticalLayoutProjectBrowser)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout()
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.labelOpenProject = QtWidgets.QLabel(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.labelOpenProject.sizePolicy().hasHeightForWidth())
        self.labelOpenProject.setSizePolicy(sizePolicy)
        self.labelOpenProject.setObjectName("labelOpenProject")
        self.verticalLayout_2.addWidget(self.labelOpenProject)
        self.btnOpenProject = QtWidgets.QPushButton(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.btnOpenProject.sizePolicy().hasHeightForWidth())
        self.btnOpenProject.setSizePolicy(sizePolicy)
        self.btnOpenProject.setMinimumSize(QtCore.QSize(131, 118))
        self.btnOpenProject.setMaximumSize(QtCore.QSize(131, 118))
        self.btnOpenProject.setAutoFillBackground(True)
        self.btnOpenProject.setText("")
        self.btnOpenProject.setFlat(False)
        self.btnOpenProject.setObjectName("btnOpenProject")
        self.verticalLayout_2.addWidget(self.btnOpenProject)
        self.horizontalLayout.addLayout(self.verticalLayout_2)
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.labelNewProj = QtWidgets.QLabel(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.labelNewProj.sizePolicy().hasHeightForWidth())
        self.labelNewProj.setSizePolicy(sizePolicy)
        self.labelNewProj.setObjectName("labelNewProj")
        self.verticalLayout.addWidget(self.labelNewProj)
        self.btnNewProject = QtWidgets.QPushButton(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.btnNewProject.sizePolicy().hasHeightForWidth())
        self.btnNewProject.setSizePolicy(sizePolicy)
        self.btnNewProject.setMinimumSize(QtCore.QSize(131, 118))
        self.btnNewProject.setMaximumSize(QtCore.QSize(131, 118))
        self.btnNewProject.setAutoFillBackground(True)
        self.btnNewProject.setText("")
        self.btnNewProject.setFlat(False)
        self.btnNewProject.setObjectName("btnNewProject")
        self.verticalLayout.addWidget(self.btnNewProject)
        self.horizontalLayout.addLayout(self.verticalLayout)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.verticalLayout_7.addLayout(self.horizontalLayout)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.verticalLayout_6 = QtWidgets.QVBoxLayout()
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.labelRecentProjects = QtWidgets.QLabel(self.centralwidget)
        self.labelRecentProjects.setObjectName("labelRecentProjects")
        self.verticalLayout_6.addWidget(self.labelRecentProjects)
        self.listWidgetRecentProjects = QtWidgets.QListWidget(self.centralwidget)
        self.listWidgetRecentProjects.setObjectName("listWidgetRecentProjects")
        self.verticalLayout_6.addWidget(self.listWidgetRecentProjects)
        self.treeViewFlowcharts = QtWidgets.QTreeView(self.centralwidget)
        self.treeViewFlowcharts.setObjectName("treeViewFlowcharts")
        self.verticalLayout_6.addWidget(self.treeViewFlowcharts)
        self.horizontalLayout_2.addLayout(self.verticalLayout_6)
        self.verticalLayout_5 = QtWidgets.QVBoxLayout()
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.labelProjectPlots = QtWidgets.QLabel(self.centralwidget)
        self.labelProjectPlots.setObjectName("labelProjectPlots")
        self.verticalLayout_5.addWidget(self.labelProjectPlots)
        self.treeViewProjectPlots = QtWidgets.QTreeView(self.centralwidget)
        self.treeViewProjectPlots.setObjectName("treeViewProjectPlots")
        self.verticalLayout_5.addWidget(self.treeViewProjectPlots)
        self.horizontalLayout_2.addLayout(self.verticalLayout_5)
        self.verticalLayout_7.addLayout(self.horizontalLayout_2)
        self.label = QtWidgets.QLabel(self.centralwidget)
        self.label.setObjectName("label")
        self.verticalLayout_7.addWidget(self.label)
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 715, 30))
        self.menubar.setObjectName("menubar")
        self.menuConfiguration = QtWidgets.QMenu(self.menubar)
        self.menuConfiguration.setObjectName("menuConfiguration")
        self.menuView = QtWidgets.QMenu(self.menubar)
        self.menuView.setObjectName("menuView")
        self.menuTools = QtWidgets.QMenu(self.menubar)
        self.menuTools.setObjectName("menuTools")
        self.menuHelp = QtWidgets.QMenu(self.menubar)
        self.menuHelp.setObjectName("menuHelp")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.actionProject_Configuration = QtWidgets.QAction(MainWindow)
        self.actionProject_Configuration.setEnabled(False)
        self.actionProject_Configuration.setObjectName("actionProject_Configuration")
        self.actionSystem_Configuration = QtWidgets.QAction(MainWindow)
        self.actionSystem_Configuration.setObjectName("actionSystem_Configuration")
        self.actionShow_all_windows = QtWidgets.QAction(MainWindow)
        self.actionShow_all_windows.setObjectName("actionShow_all_windows")
        self.actionConsole = QtWidgets.QAction(MainWindow)
        self.actionConsole.setCheckable(True)
        self.actionConsole.setObjectName("actionConsole")
        self.actionViewer_linker = QtWidgets.QAction(MainWindow)
        self.actionViewer_linker.setObjectName("actionViewer_linker")
        self.actionDocs_homepage = QtWidgets.QAction(MainWindow)
        self.actionDocs_homepage.setObjectName("actionDocs_homepage")
        self.actionNew_Project_Docs = QtWidgets.QAction(MainWindow)
        self.actionNew_Project_Docs.setObjectName("actionNew_Project_Docs")
        self.actionSystem_Configuration_Docs = QtWidgets.QAction(MainWindow)
        self.actionSystem_Configuration_Docs.setObjectName("actionSystem_Configuration_Docs")
        self.action_version = QtWidgets.QAction(MainWindow)
        self.action_version.setText("")
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.action_version.setFont(font)
        self.action_version.setObjectName("action_version")
        self.actionReport_issue_bug = QtWidgets.QAction(MainWindow)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.actionReport_issue_bug.setFont(font)
        self.actionReport_issue_bug.setObjectName("actionReport_issue_bug")
        self.actionQuestions_discussion = QtWidgets.QAction(MainWindow)
        self.actionQuestions_discussion.setObjectName("actionQuestions_discussion")
        self.menuConfiguration.addAction(self.actionProject_Configuration)
        self.menuConfiguration.addAction(self.actionSystem_Configuration)
        self.menuView.addAction(self.actionShow_all_windows)
        self.menuView.addSeparator()
        self.menuView.addAction(self.actionConsole)
        self.menuTools.addAction(self.actionViewer_linker)
        self.menuHelp.addAction(self.actionReport_issue_bug)
        self.menuHelp.addAction(self.actionQuestions_discussion)
        self.menuHelp.addSeparator()
        self.menuHelp.addAction(self.actionDocs_homepage)
        self.menuHelp.addAction(self.actionNew_Project_Docs)
        self.menuHelp.addAction(self.actionSystem_Configuration_Docs)
        self.menuHelp.addSeparator()
        self.menuHelp.addAction(self.action_version)
        self.menubar.addAction(self.menuConfiguration.menuAction())
        self.menubar.addAction(self.menuView.menuAction())
        self.menubar.addAction(self.menuTools.menuAction())
        self.menubar.addAction(self.menuHelp.menuAction())

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "MainWindow"))
        self.label_4.setText(_translate("MainWindow", "Viewer"))
        self.label_5.setText(_translate("MainWindow", "Flowchart"))
        self.labelProjectBrowser.setText(_translate("MainWindow", "Project Browser"))
        self.labelOpenProject.setText(_translate("MainWindow", "Open Project"))
        self.labelNewProj.setText(_translate("MainWindow", "New Project"))
        self.labelRecentProjects.setText(_translate("MainWindow", "Recent projects"))
        self.labelProjectPlots.setText(_translate("MainWindow", "Project plots"))
        self.label.setText(_translate("MainWindow", "Console"))
        self.menuConfiguration.setTitle(_translate("MainWindow", "Confi&guration"))
        self.menuView.setTitle(_translate("MainWindow", "&View"))
        self.menuTools.setTitle(_translate("MainWindow", "Too&ls"))
        self.menuHelp.setTitle(_translate("MainWindow", "Help"))
        self.actionProject_Configuration.setText(_translate("MainWindow", "&Project Configuration"))
        self.actionSystem_Configuration.setText(_translate("MainWindow", "&System Configuration"))
        self.actionShow_all_windows.setText(_translate("MainWindow", "&Show all windows"))
        self.actionConsole.setText(_translate("MainWindow", "&Console"))
        self.actionViewer_linker.setText(_translate("MainWindow", "&Viewer linker"))
        self.actionDocs_homepage.setText(_translate("MainWindow", "&Docs homepage"))
        self.actionNew_Project_Docs.setText(_translate("MainWindow", "&New Project help"))
        self.actionSystem_Configuration_Docs.setText(_translate("MainWindow", "&System Config help"))
        self.actionReport_issue_bug.setText(_translate("MainWindow", "Report issue/bug"))
        self.actionQuestions_discussion.setText(_translate("MainWindow", "Questions/discussion"))




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on June 14 2018

@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007

A back-end module for managing most project related functions, such as the root dataframe and all children,
adding rows to the root dataframe, updating child dataframes, and updating the project configuration.
"""

from PyQt5 import QtCore
from ..common import configuration, project_config_window, start, get_window_manager, is_mesmerize_project
# from ..common import get_window_manager
import os
import pandas as pd
from time import time
from shutil import move as move_file
from warnings import warn


class ProjectManager(QtCore.QObject):
    signal_dataframe_changed = QtCore.pyqtSignal(pd.DataFrame)
    signal_project_config_changed = QtCore.pyqtSignal()

    def __init__(self):
        QtCore.QObject.__init__(self)
        self.root_dir = None
        self.dataframe = pd.DataFrame(data=None)
        self.child_dataframes = None

    def set(self, project_root_dir: str):
        self.root_dir = project_root_dir
        configuration.proj_path = self.root_dir
        self.dataframe = pd.DataFrame(data=None)
        self.signal_dataframe_changed.connect(self.save_dataframe)
        self.child_dataframes = dict()

    def create_sub_dataframe(self):
        for child_name in configuration.proj_cfg.options('CHILD_DFS'):
            filt = configuration.proj_cfg['CHILD_DFS'][child_name]
            df = self.dataframe.copy()
            filt = filt.split('\n')
            for f in filt:
                df = eval(f)
            self.child_dataframes.update({child_name: {'dataframe': df, 'filter_history': filt}})

    def add_sub_dataframe(self, child_name: str, filter_history: str, dataframe: pd.DataFrame):
        self.child_dataframes.update({child_name: {'dataframe': dataframe, 'filter_history': filter_history}})
        if child_name in configuration.proj_cfg.options('CHILD_DFS'):
            configuration.proj_cfg['CHILD_DFS'][child_name] = configuration.proj_cfg['CHILD_DFS'][child_name] + '\n'.join(filter_history)
        else:
            configuration.proj_cfg.set('CHILD_DFS', child_name, '\n'.join(filter_history))

        configuration.save_proj_config()

    def remove_sub_dataframe(self, name: str):
        self.child_dataframes.pop(name)
        configuration.proj_cfg.remove_option('CHILD_DFS', name)
        configuration.save_proj_config()

    def setup_new_project(self):
        subdirs = ['dataframes',
                   'batches',
                   'images',
                   'flowcharts',
                   'trns',
                   'plots']

        for d in subdirs:
            os.makedirs(os.path.join(self.root_dir, d))

        configuration.create_new_proj_config()

        self._initialize_config_window()
        self.config_window.tabs.widget(0).ui.btnSave.clicked.connect(self._create_new_project_dataframe)
        self.config_window.show()

    def open_project(self):
        is_mesmerize_project(self.root_dir)

        df_path = os.path.join(self.root_dir, 'dataframes', 'root.dfr')
        self.dataframe = pd.read_hdf(df_path, key='project_dataframe', mode='r')

        self._initialize_config_window()

        configuration.open_proj_config()

    def attach_open_windows(self, windows):
        pass

    def update_open_windows(self, windows):
        pass

    def _initialize_config_window(self):
        self.config_window = project_config_window.Window()
        self.config_window.tabs.widget(0).ui.btnSave.clicked.connect(self.update_open_windows)
        # self.config_window.resize(560, 620)

    def show_config_window(self):
        self.config_window.show()

    def _create_new_project_dataframe(self):
        self.config_window.tabs.widget(0).ui.btnSave.clicked.disconnect(self._create_new_project_dataframe)

        include = configuration.proj_cfg.options('INCLUDE')
        exclude = configuration.proj_cfg.options('EXCLUDE')

        c = include + exclude

        self.dataframe = pd.DataFrame(data=None, columns=c)
        self.save_dataframe()

        start.project_browser()

    def get_dataframe(self) -> pd.DataFrame:
        return self.dataframe

    def set_dataframe(self, dataframe: pd.DataFrame):
        if not isinstance(dataframe, pd.DataFrame):
            raise TypeError('Must pass an instance of pandas.DataFrame')
        self.dataframe = dataframe.copy(deep=True)
        self.emit_signal_dataframe_changed()

    def save_dataframe(self):
        df_path = os.path.join(self.root_dir, 'dataframes', 'root.dfr')
        if os.path.isfile(df_path):
            warn('root.dfr already exists, renaming file...')
            self.backup_project_dataframe()

        self.dataframe.to_hdf(df_path, key='project_dataframe', mode='w')

    def update_project_config_requested(self, custom_to_add: dict):
        if self.dataframe.empty:
            return

        columns_changed = False

        for column in configuration.proj_cfg.options('ROI_DEFS'):
            if column not in self.dataframe.columns:
                self.dataframe[column] = 'untagged'
                columns_changed = True

        for column in configuration.proj_cfg.options('STIM_DEFS'):
            if column not in self.dataframe.columns:
                self.dataframe[column] = [['untagged']] * len(self.dataframe)
                columns_changed = True

        for column in configuration.proj_cfg.options('CUSTOM_COLUMNS'):
            if column not in self.dataframe.columns:

                rep_val = custom_to_add[column]['replacement_value']

                if custom_to_add[column]['type'] == 'bool':
                    if rep_val == 'True':
                        rep_val = True
                    elif rep_val == 'False':
                        rep_val = False

                self.dataframe[column] = rep_val
                columns_changed = True

        all_columns = configuration.proj_cfg.options('INCLUDE') + configuration.proj_cfg.options('EXCLUDE')
        columns_to_drop = list(set(self.dataframe.columns).difference(all_columns))
        if len(columns_to_drop) > 0:
            columns_changed = True

        if columns_changed:
            self.backup_project_dataframe()
            self.dataframe.drop(columns=columns_to_drop, inplace=True)
            self.save_dataframe()

            self.signal_dataframe_changed.disconnect(get_window_manager().project_browser.project_browser.update_dataframe_data)
            get_window_manager().project_browser.deleteLater()

            self.signal_project_config_changed.emit()
            self.emit_signal_dataframe_changed()

            pb = start.project_browser()
            pb.reload_all_tabs()
            get_window_manager().project_browser = pb

        # configuration.proj_cfg_changed.notify_all()

    def delete_gui_children(self, widget: QtCore.QObject):
        widget.children()

    def backup_project_dataframe(self):
        move_file(os.path.join(self.root_dir, 'dataframes', 'root.dfr'), os.path.join(self.root_dir, 'dataframes', f'root_bak_{time()}.dfr'))

    def append_to_dataframe(self, dicts_to_append: list):
        self.backup_project_dataframe()
        self.dataframe = self.dataframe.append(pd.DataFrame(dicts_to_append), ignore_index=True)
        self.emit_signal_dataframe_changed()

    def emit_signal_dataframe_changed(self):
        self.signal_dataframe_changed.emit(self.dataframe)

    def change_sample_rows(self, sample_id: str, dicts_to_append: list):
        """
        Remove the rows corresponding to the passed sample_id and replace them with the list of dicts provided
        """
        # self.backup_project_dataframe()
        # self.dataframe = self.dataframe[self.dataframe['SampleID'] != sample_id]
        self.delete_sample_id_rows(sample_id)
        self.dataframe = self.dataframe.append(pd.DataFrame(dicts_to_append), ignore_index=True)
        self.emit_signal_dataframe_changed()

    def delete_sample_id_rows(self, sample_id: str):
        self.backup_project_dataframe()
        self.dataframe = self.dataframe[self.dataframe['SampleID'] != sample_id]
        self.emit_signal_dataframe_changed()

    def get_sample_id_rows(self, sample_id: str) -> pd.DataFrame:
        return self.dataframe[self.dataframe['SampleID'] == sample_id]
